package WMC;

use WMC_CONSTANTS;
## GENERAL FUNCTIONS
############################
sub From_PDB_ATOMS_To_Seq 
{
    my $Protein_PDB_File=$_[0];
    my $Protein_Chain_ID=$_[1];
    
    my $PDB_ID;
    if ($Protein_PDB_File=~/([A-Za-z0-9]{4}).ent/)
    {
	$PDB_ID=uc($1);
    }
    my %AA_Abb=(GLY=>'G',ALA=>'A',VAL=>'V',LEU=>'L',ILE=>'I',MET=>'M',PHE=>'F',TRP=>'W',PRO=>'P',SER=>'S',THR=>'T',CYS=>'C',TYR=>'Y',ASN=>'N',GLN=>'Q',ASP=>'D',GLU=>'E',LYS=>'K',ARG=>'R',HIS=>'H',);
    my $Protein_Name;
    if ($Protein_PDB_File=~/([1-9a-zA-Z.]+)$/)
    {
	$Protein_Name=$1;
    }
    my $first=0; #untile the first AA is found $first==0
    my $Prev_AA=0;
    my @Atoms_Seq;
    open (Protein_PDB,$Protein_PDB_File) || die "Can't open pdb file\n";
    
    while (my $line=<Protein_PDB>)
    {
	my $AA_Name;
	if (($line=~/^ATOM/)&&($first==0))
	{
	    $first=1;
	    my $First_PDB_AA;
	    if ($line=~ /^.{21}([A-Z ]{1})([0-9 ]{4})/)
	    {
		$First_PDB_AA=$2;
		$First_PDB_AA=remove_whitespace($First_PDB_AA);
	    }
	}
	if ($line=~/^ATOM/)
	{
	    my $AA_Index;
	    my $PDB_Chain_ID;
	    if ($line=~ /^.{17}([A-Z]{3}).{1}([A-Z0-9 ]{1})([0-9 ]{4})/)
	    {
		$PDB_Chain_ID=$2;
		$AA_Index=$3;
		$AA_Name=$1;
		$AA_Name=remove_whitespace($AA_Name);
		$AA_Index=remove_whitespace($AA_Index);
		$PDB_Chain_ID=remove_whitespace($PDB_Chain_ID);
		if ($PDB_Chain_ID eq "")
		{
		    $PDB_Chain_ID=0;
		}
	    }
	    if (($AA_Index>$Prev_AA)&&($AA_Index ne "")&&($PDB_Chain_ID eq $Protein_Chain_ID))
	    {
		$Atoms_Seq[$AA_Index]=$AA_Abb{$AA_Name};
		$Prev_AA=$AA_Index;
	    }
	}
    }
    my $Seq=">$PDB_ID"."_$Protein_Chain_ID ATOMS\n";
    my $i=1;
    my $Seq_Length=@Atoms_Seq;
    for ($i=1;$i<$Seq_Length;$i++)
    {
	if (! defined $Atoms_Seq[$i])
	{
	    $Atoms_Seq[$i]="X";
	}
    }
    close (Protein_PDB);
    #   print "SEQ:$Seq","JOIN:",join ("",@Atoms_Seq[1..($Seq_Length-1)]),"\n";<STDIN>;
    $Seq=$Seq.join ("",@Atoms_Seq[1..($Seq_Length-1)])."\n";
    return $Seq;
}

sub remove_whitespace($)
{
  my $string = shift;
  $string =~ s/^\s+//;
  $string =~ s/\s+$//;
  return $string;
}

sub Analyze_PSI_Blast 
  {
    my $PSI_Blast_Output_File=$_[0];
    my $Q_LowerLimitOnAlignmentLength=$_[1];
    my $Min_AlignmentLength=$_[2];
    my $S_LowerLimitOnAlignmentLength=$_[3];
    my $minID=$_[4];
    my $ProteinPerOrg=$_[5];
    my $Q_UpperLimitOnAlignmentLength=$_[6];
    my $S_UpperLimitOnAlignmentLength=$_[7];
    my $maxID=$_[8];
    my $Round=$_[9];
    my $min_E_Value=$_[10];
    my $maxSIM=$_[11];
    my $E_Value_Cutoff_Flag=$_[12];
    my $ID_Cutoff_Flag=$_[13];
    my $Method=$_[14];
    my ($QueryGI,$QueryLength);
    my %PDB_Taken=();
    my %Organisms=();
    my @PDB_IDs=();
    my $Analyze_Report_File="$PSI_Blast_Output_File".".Anlyze_Report";
    $Analyze_Report_File=$Analyze_Report_File.".Min_Align$Min_AlignmentLength" if ($Min_AlignmentLength>-1);
    
    print "Analyze_Report_File:$Analyze_Report_File\n";
    open (ERROR,">$Analyze_Report_File") || die "Can't open Analyze Report File: '$Analyze_Report_File' $!";
    open (PSI_BLAST,$PSI_Blast_Output_File) || die "Cna't Open PSI-Blast File: '$PSI_Blast_Output_File' $!";
    my $line=<PSI_BLAST>;
    while ($line and ($line!~/^Query\= (T[0-9]+)/) and ($line!~/^Query\= gi\|(\d+_*\d*)\|/) and $line!~/^Query\= ([A-Z0-9]+_[A-Z0-9]+)/) {$line=<PSI_BLAST>;}
    if ($line and (($line=~/^Query\= gi\|(\d+_*\d*)\|/) or ($line=~/^Query\= (T[0-9]+)/) or ($line=~/^Query\= ([A-Z0-9]+_[A-Z0-9_]+)/) or ($line=~/^Query\= ([A-Z0-9a-z_]+)/)))
    {
	chomp($line);
	my $Annotation=$line;
	while ($line and ($line!~/\((\d+)\s+letters/)) {$line=<PSI_BLAST>; chomp($line); $Annotation.=" $line";}
	
	$Annotation=~s/\s+/ /g;
	if( ($Annotation =~ /^Query\= gi\|(\d+_*\d*)\|\[(.+)\]\|.*\((\d+) letters\)/) or 
	    ($Annotation =~ /^Query\= gi\|(\d+_*\d*)\|\s*(\S*)\s*\|.*\((\d+) letters\)/))
	{
	    $QueryGI=$1;
	    my $orgn=$2; 
	    print "org $orgn\n";
	    $QueryLength=$3;
	    $Organisms{$orgn}++;
	}
	elsif (($Annotation =~ /^Query\= (T[0-9]+).*\((\d+) letters\)/)) # FOR CASP TARGETS FORMAT
	{
	    $QueryGI=$1;
	    $QueryLength=$2;
	}
	elsif (($Annotation =~ /^Query\= ([A-Z0-9]+_[A-Z0-9_]+).*\((\d+) letters\)/)) # FOR SP TARGETS FORMAT
	{
	    $QueryGI=$1;
	    $QueryLength=$2;
	}
	elsif (($Annotation =~ /^Query\= ([A-Z0-9a-z_]+).*\((\d+) letters\)/)) # FOR OTHER TARGETS FORMAT
	{
	    $QueryGI=$1;
	    $QueryLength=$2;
	}
	else 
	{
	    print ERROR "Query Expected in Annotation=$Annotation\n"; 
	    return 0;
	}
    }
    else 
    {
	print ERROR "Query Expected in $line\n";
    }
    
    if ($Round)
    {
	while ($line and ($line!~/Results from round $Round/)) {$line=<PSI_BLAST>;}
    }
    
    while (my $line=<PSI_BLAST>)
    {
	last if ($line =~ /Number of sequences in database/);
	while ($line and ($line!~/^\>/)&&($line !~ /Number of sequences in database/)) {$line=<PSI_BLAST>;}
	last if ($line =~ /Number of sequences in database/);
	
	chomp($line);
	my $Annotation=$line;
	while ($line and ($line!~/Length \= /)) {$line=<PSI_BLAST>; chomp($line); $Annotation.=$line;}
	$Annotation=~s/\s+/ /g;
	# >1AWQA 164 XRAY 1.58 0.343 0.433 no CYCLOPHILIN A <UNP PPIA_HUMAN> [HOMO SAPIENS] Length = 164
	# >1IHGA 370 XRAY 1.80 0.178 0.256 no Cyclophilin 40 <UNP PPID_BOVIN> [BOS TAURUS] Length = 370
	# >gi|51588442_0|[Yersinia pseudo tuberculosis IP 32953]| ORIGINAL gi|51588442|emb|CAH20050.1| putative cytochrome b(562) [Yersinia pseudotuberculosis IP 32953] gi|51595160|ref|YP_069351.1| putative cytochrome b(562) [Yersinia pseudotuberculosis IP 32953] Length = 127*
	if ($Annotation =~ /^\>([A-Za-z0-9_]+)\s+(.*)\<(.+)\>\s+\[(.+)\].*Length\s+=\s+(\d+)/)
	  {
	      # Extract the organism from [] put 1 in $Organisms{$3} and consider only the first hit of every organism. Some SP do not have [], but only _organism, so at the second stage make sure _HUMAN=[Homo Sapiens].
	      my $Paralog=0;
	      my $PDB_ID=$1;
	      my @Annotation=split (/\s+/,$2);
	      my $PDB_Reference_Seq_Name=$3;
	      my $orgn=$4; 
	      my $SubjectLength=$5;
	      while ($line and $line!~/Score \=/) {$line=<PSI_BLAST>;}
	      chomp ($line);
	      my $line1=<PSI_BLAST>;
	      $line=$line.$line1;
	      if ($line =~ /Identities \= \d+\/\d+ \((\d+)\%\)\, Positives \= \d+\/\d+ \((\d+)\%\)/) 
	      {
		  my $ID=$1;
		  my $Sim=$2;
		  my $E_Value=$1 if ($line =~/Expect\s+=\s+([e0-9.\-]+)/);
		  $E_Value="1".$E_Value if ($E_Value=~/^e-/);
		  my ($Q_AlignmentLength,$S_AlignmentLength); 
		  my ($Query,$Sbjct);
		  while ($line and ($line!~/^Query/)) {$line=<PSI_BLAST>;}
		  while ($line =~ /^Query/)
		  {
		      chomp($line);
		      $Query.=$line;
		      $line=<PSI_BLAST>;
		      $line=<PSI_BLAST>;
		      chomp($line);
		      $Sbjct.=$line;
		      $line=<PSI_BLAST>; 
		      $line=<PSI_BLAST>;
		  }
		  
		  if ($Query=~/^Query: (\d+)\s+.+\s+(\d+)$/)
		  {
		      my $Start=$1; my $End=$2;
		      $Q_AlignmentLength=$End-$Start+1;
		  }
		  else {print ERROR "Problem in Query Format of $PDB_ID: $Query";}
		
		  if ($Sbjct=~/^Sbjct: (\d+)\s+.+\s+(\d+)$/)
		  {
		      my $Start=$1; my $End=$2;
		      $S_AlignmentLength=$End-$Start+1;
		  }
		  else {print ERROR "Problem in Sbjct Format of $PDB_ID: $Sbjct";}
		  if (defined $Organisms{$orgn})
		  {
		      if ($Organisms{$orgn}>=$ProteinPerOrg)
		      {
			  $Paralog=1;
		      }
		  }
		  # MIN ALIGNMENT LENGTH=$Q_LowerLimitOnAlignmentLength
		  ###############################
		  if ($Min_AlignmentLength==-1)
		  {
		      $Min_AlignmentLength=$QueryLength*$Q_LowerLimitOnAlignmentLength;
		  }
		  #####################################################################
		  if ($PDB_ID eq $QueryGI)
		  {
		      print ERROR "PDB_ID=$PDB_ID is the Query, so it is already there\n";
		  }
		  elsif (($ID>$maxID and !$Include_Very_Close))
		  {print ERROR "PDB_ID=$PDB_ID ID is $ID while Proteins With More Than $maxID Not Included\n";}
		  elsif ($ID<$minID and $ID_Cutoff_Flag)
		  {print ERROR "PDB_ID=$PDB_ID has only $ID percent identity\n";}
		  elsif ($E_Value>$min_E_Value and $E_Value_Cutoff_Flag)
		  {print ERROR "PDB_ID=$PDB_ID has E-Value of $E_Value while threshold is $min_E_Value\n";}
		  elsif ($Sim>$maxSIM and !$Include_Very_Close)
		  {print ERROR "PDB_ID=$PDB_ID was not taken becuse %Similarity $Sim>$maxSIM\n";}
		  elsif (($Q_AlignmentLength/$QueryLength < $Q_LowerLimitOnAlignmentLength)&&($Q_AlignmentLength<$Min_AlignmentLength))
		  {print ERROR "PDB_ID=$PDB_ID was not used because short alignment on Query: Q_AlignmentLength=$Q_AlignmentLength QueryLength=$QueryLength Q_LowerLimitOnAlignmentLength=$Q_LowerLimitOnAlignmentLength\n";}
		  elsif ($Q_AlignmentLength/$QueryLength > $Q_UpperLimitOnAlignmentLength)
		  {print ERROR "PDB_ID=$PDB_ID was not used because long alignment on Query: Q_AlignmentLength=$Q_AlignmentLength QueryLength=$QueryLength Q_UpperLimitOnAlignmentLength=$Q_UpperLimitOnAlignmentLength\n";}
		  elsif ($S_AlignmentLength/$SubjectLength < $S_LowerLimitOnAlignmentLength)
		  {print ERROR "PDB_ID=$PDB_ID was not used because short alignment on Subject: S_AlignmentLength=$S_AlignmentLength SubjectLength=$SubjectLength S_LowerLimitOnAlignmentLength=$S_LowerLimitOnAlignmentLength\n";}
		  elsif ($S_AlignmentLength/$SubjectLength > $S_UpperLimitOnAlignmentLength)
		  {print ERROR "PDB_ID=$PDB_ID was not used because long alignment on Subject: S_AlignmentLength=$S_AlignmentLength SubjectLength=$SubjectLength S_UpperLimitOnAlignmentLength=$S_LowerLimitOnAlignmentLength\n";}
		  elsif ($Paralog==1)
		  {print ERROR "PDB=$PDB_ID was not used because $orgn was already used \n";}
		  elsif ($Annotation[1] ne $Method and $Method ne "All")
		  {print ERROR "PDB=$PDB_ID was not used because Only $Method Sructure Was Used while its $Annotation[1] Structure\n";}
		  else 
		  {
		      $PDB_Taken{$PDB_ID}{'ID'}=$ID;
		      $PDB_Taken{$PDB_ID}{'Q_AlignmentLength'}=$Q_AlignmentLength;
		      $PDB_Taken{$PDB_ID}{'S_AlignmentLength'}=$S_AlignmentLength;
		      $PDB_Taken{$PDB_ID}{'Organism'}=$orgn;
		      $PDB_Taken{$PDB_ID}{'Length'}=$Annotation[0];
		      $PDB_Taken{$PDB_ID}{'Method'}=$Annotation[1];
		      $PDB_Taken{$PDB_ID}{'Resolution'}=$Annotation[2];
		      $PDB_Taken{$PDB_ID}{'Ref_Seq_Name'}=$PDB_Reference_Seq_Name;
		      $PDB_Taken{$PDB_ID}{'E_Value'}=$E_Value;
		      $PDB_Taken{$PDB_ID}{'SIM'}=$Sim;
		      push (@PDB_IDs,$PDB_ID);
		      $Organisms{$orgn}++;
		  }
	      }
	      else 
	      {
		  print ERROR "Problem in File format. Identities expected in $line*\n";
	      }
	  }
	
    }
    return (\@PDB_IDs,\%PDB_Taken);
}

sub How_Many_Psi_Rounds
{
    my $PSI_Blast_Out=$_[0];
    my $Round=1;
    open (PSI_BLAST_OUTPUT,$PSI_Blast_Out) || die "Can't Open PSI-Blast Output: '$PSI_Blast_Out' $!";
    while (my $line=<PSI_BLAST_OUTPUT>)
    {
	chomp ($line);
	if ($line=~/^Results from round ([0-9]+)/)
	{
	    $Round=$1;
	}
    }
    close (PSI_BLAST_OUTPUT);
    return $Round;
}

sub From_Tree_to_Distance_Matrix
  {
    # Parse in newick/new hampshire tree format using BioPerl Bio::TreeIO
    my $Tree_File=$_[0];
    my $Distance_Matrix_File=$_[1];
    my $Target_Seq_Name=$_[2];
    
    open (OUT,">$Distance_Matrix_File") || die "Can't Open Distance Matrix File For Writing '$Distance_Matrix_File' $!";
    # parse in newick/new hampshire format using BioPerl Bio::TreeIO
    my $input = new Bio::TreeIO(-file   => $Tree_File,
				-format => "newick");
    my $tree = $input->next_tree;
    
    my $root = $tree->get_root_node;
    my @leaves = $tree->get_leaf_nodes;

    my @lvs_dists = ();
    my %distHash = ();
    my $total_leaves = scalar(@leaves);
    my $dist;
    foreach my $leaf1 (@leaves) 
      {
	foreach my $leaf2 (@leaves)
	  {
	    $dist = $tree->distance(-nodes => [$leaf1,$leaf2]);
	    $distHash{"$leaf1"}{"$leaf2"} = $dist; # store using the allocation name as a string (as ids may repeat)
	    if ($leaf1->id eq $Target_Seq_Name)
	      {
		print OUT $leaf1->id,"\t",$leaf2->id,"\t",$dist,"\n"; 
	      }
	  }
      }
    close (OUT);
  }


sub Compare_Dates # Returns 1 if the first date is greater than or equal to second date and 0 otherwise
  {
    my $Date1=uc $_[0];
    my $Date2=uc $_[1];
    
    my %Months=(JAN=>'1',
		FEB=>'2',
		MAR=>'3',
		APR=>'4',
		MAY=>'5',
		JUN=>'6',
		JUL=>'7',
		AUG=>'8',
		SEP=>'9',
		OCT=>'10',
		NOV=>'11',
		DEC=>'12');
#    uc($Date1);
#    uc($Date2);
    my @Date1;
    my @Date2;
    if ($Date1=~/\//)
      {
	@Date1=split ("/",$Date1);
      }
    elsif ($Date1=~/-/)
      {
	@Date1=split ("-",$Date1);
      }
    else
      {
	die "ERROR: Date is $Date1 while Only Dates In Format day/month/year or day-month-year is acceptable";
      }
    if ($Date2=~/\//)
      {
	@Date2=split ("/",$Date2);
      }
    elsif ($Date2=~/-/)
      {
	@Date2=split ("-",$Date2);
      }
    else
      {
	die "ERROR: Date is $Date2 while Only Dates In Format day/month/year or day-month-year is acceptable";
      }
    my $year1=$Date1[2];
    if (($year1=~/^[01]/) and (length($year1)==2))
      {
	$year1="20".$year1;
      }
    elsif (($year1!~/^0/) and (length($year1)==2))
      {
	$year1="19".$year1;
      }
    my $year2=$Date2[2];
    if (($year2=~/^[01]/) and (length ($year2)==2))
      {
	$year2="20".$year2;
      }
    elsif (($year2!~/^0/) and (length($year2)==2))
      {
	$year2="19".$year2;
      }
    my $month1=$Date1[1];
    if (defined $Months{$month1})
    {
	$month1=$Months{$month1};
    }
    my $month2=$Date2[1];
    if (defined $Months{$month2})
    {
	$month2=$Months{$month2};
    }
    my $day1=$Date1[0];
    my $day2=$Date2[0];
    if ($year1>$year2)
    {
	return 1;
    }
    elsif ($year2>$year1)
    {
	return 0;
    }
    elsif ($year2==$year1)
    {
	if ($month1>$month2)
	{
	    return 1;
	}
	elsif ($month2>$month1)
	{
	    return 0;
	}
	elsif ($month2==$month1)
	{
	    if ($day1>$day2)
	    {
		return 1;
	    }
	    elsif ($day2>$day1)
	    {
		return 0;
	    }
	    elsif ($day2==$day1)
	    {
		return 1;
	    }
	}
    }
}
sub PDB_Exist
{
    my $PDB_ID=$_[0];
    my $PDB_ID_Flag=0;
    my $PDB_Exist=0;
    my $Tmp_PDB_File;
    my $PDB_Not_Found=0;
    if (defined $PDB_ID)
    {
	my $PDB_ID_Name=lc($PDB_ID);
	my $PDB_File=$PDB_Path.substr($PDB_ID_Name, 1,2)."/pdb".substr($PDB_ID_Name, 0,4).".ent.gz";
	if (-e $PDB_File)
	{
	    return 1;
	}
	else
	{
	    return 0;
	}
    }
}

## CSU CONTACT MAP
######################
### Will run the CSU algorithm and create a contactMap 
sub CSU_CONTACT_MAP
{
    my $PDB_ID=shift;
    my $Out_Path=shift;
    my $Tmp_Path=shift;
    
#PDB Location
#=============
    my $PDB_Path=WMC_CONSTANTS::PDB_LOCAL_PATH;
    my $WGET_PATH=WMC_CONSTANTS::PDB_WGET_PATH;
    
#Chek For PDB ID
#=====================
    my $PDB_ID_Flag=0;
    my $PDB_Exist=0;
    my $Tmp_PDB_File;
    my $PDB_ID_Name=lc($PDB_ID);
    my $PDB_File=$PDB_Path.substr($PDB_ID_Name, 1,2)."/pdb".substr($PDB_ID_Name, 0,4).".ent.gz";
    $Tmp_PDB_File=$Tmp_Path.substr($PDB_ID_Name, 0,4).".ent";
    if (-e  $PDB_File)
    {
	system ("zcat $PDB_File | grep -v \"ANISOU\" > $Tmp_PDB_File");
	$PDB_Exist=1;
    }
    else 
    {
	my $Downloaded_File=$MissingPDB_Path."pdb".substr($PDB_ID_Name, 0,4).".ent.gz";
	system("wget $WGET_PATH".substr($PDB_ID_Name, 1,2)."/pdb".substr($PDB_ID_Name, 0,4).".ent.gz -O $Downloaded_File");
	if (-e $Downloaded_File)
	{
	    system ("zcat $Downloaded_File | grep -v \"ANISOU\" > $Tmp_PDB_File"); 
	}
    }
    
    if (!-e $Tmp_PDB_File or -z $Tmp_PDB_File)
    {
	print "PDB WAS NOT FOUND: $PDB_File\n";
	exit (1);
    }
    $Protein_File_Name=$Tmp_PDB_File;
    $OutputFile_Sur=$Out_Path.$PDB_ID.".CSU";
    $Chain_ID=substr($PDB_ID, 5,5);
    open (PDB_FILE,$Protein_File_Name)||die "cannot open file: '$Protein_File_Name' $!";
    open (OUTPUT, ">$OutputFile_Sur") || die "Can't open output file: $OutputFile_Sur $!.\n";
    my $Residue_num;#=1;
    
    my $Prev_AA=0;
    my $Prev_Chain;
    my %Inter_Chain_Interactions=();
    my %Intra_Chain_Interactions=();
    my %Both_Inter_Intra_Chain_Interactions=();
    my $Interacted_AA_Name;
    my $Interacted_AA_Num;
    my $Interacted_AA_Dis;
    my $Intearcted_AA_Surf;
    my $Interacted_AA_Chain;
    my $Chain_num;
    my %AA_Names=(GLY=>'1',ALA=>'1',VAL=>'1',LEU=>'1',ILE=>'1',MET=>'1',PHE=>'1',TRP=>'1',PRO=>'1',SER=>'1',THR=>'1',CYS=>'1',
		  TYR=>'1',ASN=>'1',GLN=>'1',ASP=>'1',GLU=>'1',LYS=>'1',ARG=>'1',HIS=>'1');
    my %AA_Abbreviations=(GLY=>'G',ALA=>'A',VAL=>'V',LEU=>'L',ILE=>'I',MET=>'M',PHE=>'F',TRP=>'W',PRO=>'P',SER=>'S',THR=>'T',CYS=>'C',TYR=>'Y',ASN=>'N',GLN=>'Q',ASP=>'D',GLU=>'E',LYS=>'K',ARG=>'R',HIS=>'H');
    print OUTPUT "Intra Chain Interactions Of Chain '$Chain_ID' at: $Protein_File_Name\n\n";
    print OUTPUT "Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance Residue_A_Name Residue_B_Name\n";
    
###Getting the information about $Total_Residues and $Chain_num from The PDB File  ###
    while (my $line=<PDB_FILE>)
    {
	if ($line=~ /^ATOM/)
	{
	    if ($line=~ /^.{21}([A-Z ]{1})([0-9 ]{4})/)
	    {
		$Chain_num=$1;
		$Residue_num=$2;
	    }
	    $Chain_num=remove_whitespace($Chain_num);
	    $Residue_num=remove_whitespace($Residue_num);
### Calling CSU for the Residue of specific Chain in order to find who interacts with him
	    if (($Residue_num>$Prev_AA)||($Chain_num gt $Prev_Chain))
	    {
		my $CSU_Command=WMC_CONSTANTS::CSU." $Protein_File_Name $Residue_num $Chain_num |";
		open (README, $CSU_Command) or die "Can't Run ProgramL $!\n";
		my $Beg_Flag=0;
		my $Data;
		my $AA1_Name="";
		while (<README>)
		{
		    if ($_=~ /Residues in contact with ([A-Z]{3})/)
		    {
			$AA1_Name=$1;
		    }
		    if ($_=~m/DC$/)
		    {
			$Beg_Flag=1;
		    }
		    if (($Beg_Flag==1)&&($_=~m/^------------------/))
		    {
			$Beg_Flag=2;
		    }
		    if ($_=~ m/Table III$/)
		    {
			$Beg_Flag=0;
		    }
		    if ($Beg_Flag==2)
		    {
			$Data=1;
			if (($_=~m/^------------------/)||($_=~m/\t+$/))
			{
			    $Data=0;
			}
			if ($Data==1)
			{
			    my @line_array=split(/\s+/,$_);
			    chomp ($_);
			    if ($_ ne "")
			    {
				$Interacted_AA_Name=$line_array[2];
				$Interacted_AA_Num=$line_array[1];
				$Interacted_AA_Dis=$line_array[3];
				$Intearcted_AA_Surf=$line_array[4];
				if ($line_array[1]=~m/^(\d+)/)
				{
				    $Interacted_AA_Num=$1;
				    $Interacted_AA_Chain="";
				}
				if ($line_array[1]=~m/^(\d+)([A-Z])/)
				{
				    $Interacted_AA_Num=$1;
				    $Interacted_AA_Chain=$2;
				}
				if  ($Interacted_AA_Name=~/([A-Z]*)\*/)
				{
				    $Interacted_AA_Name=$1;
				}
				my $Pair=$Residue_num."_".$Interacted_AA_Num;
				my $Chain_num_For_File;
				if ($Chain_num eq "")
				{
				    $Chain_num_For_File=0;
				}
				else
				{
				    $Chain_num_For_File=$Chain_num;
				}
				my $Interacted_AA_Chain_For_File;
				if ($Interacted_AA_Chain eq "")
				{
				    $Interacted_AA_Chain_For_File=0;
				}
				else
				{
				    $Interacted_AA_Chain_For_File=$Interacted_AA_Chain;
				}
				
				my $Conatact_Map_Data;
				if (($AA_Abbreviations{$AA1_Name})&&($AA_Abbreviations{$Interacted_AA_Name})) # Only Natural AA
				{
				    $Conatact_Map_Data="$Residue_num $Interacted_AA_Num $Chain_num_For_File $Interacted_AA_Chain_For_File $Intearcted_AA_Surf $Interacted_AA_Dis $AA_Abbreviations{$AA1_Name} $AA_Abbreviations{$Interacted_AA_Name}"; 
				}
				if (($AA_Names{$Interacted_AA_Name}&&($Interacted_AA_Chain eq $Chain_num)&&(($Interacted_AA_Chain eq $Chain_ID)||(($Interacted_AA_Chain eq "")&&($Chain_ID eq "0"))||(($Interacted_AA_Chain eq "A")&&($Chain_ID eq "0"))))) # Choosing only those who are AA Interaction and are IntraChain interactions
				{
				    print OUTPUT "$Conatact_Map_Data\n";
				}
			    }
			    
			}
		    }
		}
	    }
	    close (README);
	    $Prev_AA=$Residue_num;
	    $Prev_Chain=$Chain_num;
	}
    }
    close (OUTPUT);
    close (PDB_FILE);
    unlink ($Protein_File_Name);
}

# Templates_Consensus_Contacts_Predictor_With_Tree_Topology.pl
sub Predict_WMC
{
    my ($ContactMap_List,$MSA,$Consensus_ContactMap,$Debug_Mode,$Only_Overlap_Mode,$Distance_Matrix,$Entropy_Like,$PhyloGenetic_Mode,$PhyloGenetic_Sum_Mode,$Tree,$Power);
    
#    PrmArgv ("Contact_Maps_List= %s ! File With Contacts Map List Of Templates", \$ContactMap_List,
#	     "MSA= %s ! Targets and Templates MSA", \$MSA,
#	     "Out= %s ! Consensus Contact Map According To the Homologs", \$Consensus_ContactMap,
#	     "Distances= %s ! Sequences Distance Matrix Compered to Target Protein - OPTIONAL for Weighted Average", \$Distance_Matrix,
#	     "Tree= %s ! The Evolutionary Tree (NEWICK format)", \$Tree,
#	     "-Debug ~ ! Flag For Printing Debug Info", \$Debug_Mode,
#	     "-Only_Overlap ~ ! Flag For considering only the overlap pairs in the averaging not all templates", \$Only_Overlap_Mode,
#	     "-Entropy_Like ~ ! Flag For considering the amount of information of the paired column", \$Entropy_Like,
#	     "-Phylogenetic_Mode ~ ! Flag For Consider Phylogenetic (Whighted Average) Information of the paired column", \$PhyloGenetic_Mode,
#	     "-Phylogenetic_Sum_Mode ~ ! Flag For Consider Phylogenetic (Whighted Sum) Information of the paired column", \$PhyloGenetic_Sum_Mode,
#	     "Power=1 %s ! Power of distance", \$Power,
#	     );
    $ContactMap_List=shift;
    $MSA=shift;
    $Consensus_ContactMap=shift;
    $Distance_Matrix=shift;
    $Tree=shift;
    $Power=shift;
    $Only_Overlap_Mode=shift;
    $Entropy_Like=shift;
    $PhyloGenetic_Mode=shift;
    $PhyloGenetic_Sum_Mode=shift;
    $Debug_Mode=shift;
    my $MINIMAL_SEPARATION=6;
#####        CONTACT MAP FORMAT     ###########
###############################################
#Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance Residue_A_Name Residue_B_Name
#70 18 0 0 1.1 6.1 T E
#135 111 0 0 13.0 4.1 L I
    
    my $ContactMap_Pos1_Col=0;
    my $ContactMap_Pos1_Chain_Col=2;
    my $ContactMap_Pos1_AA_Col=6;
    my $ContactMap_Pos2_Col=1;
    my $ContactMap_Pos2_Chain_Col=3;
    my $ContactMap_Pos2_AA_Col=7;
    my $Contact_Map_Distance_Col=5;
    
################################################
    my %DistanceMatrix=();
    
    my $Reference_Seq_ID;
    my %Consensus_Contact_Map=();
    open (MSA,$MSA) || die "Can't Open MSA: '$MSA' $!";
    $Reference_Seq_ID=<MSA>;
    chomp ($Reference_Seq_ID);
    close (MSA);
    my @Homologs_Structures;
    my $MSA_Hash_Ref=read_MSA_To_Hash_of_Predict_WMC($MSA);
    my %MSA=%$MSA_Hash_Ref;
    print "=== Finished Reading MSA: $MSA To Hash\n";
    my $Tmp=join("",@{$MSA{$Reference_Seq_ID}});
    $Tmp =~ s/-//g;
    my $Reference_Seq=$Tmp;
    my @Reference_Seq_Array=("START_FROM1",split(//,$Reference_Seq));
    print "Reference Seq: $Tmp\n";
    my $Reference_Pos_MSA_Pos_Ref=Build_Refference_Seq_MSA_Pos_Hash_of_Predict_WMC ($Reference_Seq_ID,$MSA_Hash_Ref);
    my %Reference_Pos_MSA_Pos=%$Reference_Pos_MSA_Pos_Ref;
    my $MSA_Pos_to_Without_Gap_Pos=Create_MSA_Pos_to_Without_Gap_Pos_of_Predict_WMC($MSA_Hash_Ref);
    my %MSA_Pos_to_Without_Gap_Pos=%$MSA_Pos_to_Without_Gap_Pos;
    
    my %Contact_Maps_Hash=();
    my $Contact_Maps_Hash_Ref;
    my $Target_Name;
    if ($Reference_Seq_ID =~/\>(T[0-9]+).*/)
    {
	$Target_Name=$1;
    }
    elsif ($Reference_Seq_ID =~/\>(TARGET_[0-9_]+)/)
    {
	$Target_Name=$1;
    }
    elsif ($Reference_Seq_ID =~/\>([A-Z0-9]+_[A-Z0-9]+)/)
    {
	$Target_Name=$1;
    }
    
    if (defined $Distance_Matrix)
    {
	open (DISTANCES,$Distance_Matrix) || die "Can't Open Distance Matrix File '$Distance_Matrix' $!";
	while (my $line=<DISTANCES>)
	{
	    my @line=split(/\t/,$line);
	    if ($line[1] ne $line[0])
	    {
		my $Template_PDB;
		if ($line[1]=~/([0-9A-Za-z]{4}_[A-Za-z0-9])_ATOMS/)
		{
		    $Template_PDB=$1;
		}
		elsif ($line[1]=~/([0-9A-Za-z]{4}_[A-Za-z0-9]) ATOMS/)
		{
		    $Template_PDB=$1;
		}
		my $Distance=$line[2];
		$DistanceMatrix{$Reference_Seq_ID}{$Template_PDB}=$Distance if ($Power==1);
		$DistanceMatrix{$Reference_Seq_ID}{$Template_PDB}=$Distance**$Power if ($Power>1);
		if ($DistanceMatrix{$Reference_Seq_ID}{$Template_PDB}) {$DistanceMatrix{$Reference_Seq_ID}{$Template_PDB}=0.000001;}
	    }
	}
    }
    open (CONATACTMAP_LIST,$ContactMap_List) || die "Can't Open Contact Map List: '$ContactMap_List' $!";
    while (my $ContactMap_File=<CONATACTMAP_LIST>)
    {
	my $File_Without_Path;
	my $PDB_ID;
	$Contact_Maps_Hash_Ref=read_Contact_Map_To_Hash_of_Predict_WMC($ContactMap_File,\%Contact_Maps_Hash);
	%Contact_Maps_Hash=%$Contact_Maps_Hash_Ref;
	if ($ContactMap_File=~/([A-Za-z0-9._]+)$/)
	{
	    $File_Without_Path=$1;
	    if ($File_Without_Path=~/(.*).CSU_Contact_Map/)
	    {
	    $PDB_ID=$1;
	}
	    push (@Homologs_Structures,$PDB_ID);
	}
    }
    %Contact_Maps_Hash=%$Contact_Maps_Hash_Ref;
    
    my $Num_Of_Homologs=@Homologs_Structures;
    my $Reference_Seq_Size=@Reference_Seq_Array;
    if ($Entropy_Like) # CONSIDER THE AMOUNT OF INFORMATION IN EACH PAIRED COLUMN
    {
	# COUNTING PAIRS
	for (my $Pos1=1;$Pos1<$Reference_Seq_Size;$Pos1++)
	{
	    for (my $Pos2=$Pos1+$MINIMAL_SEPARATION;$Pos2<$Reference_Seq_Size;$Pos2++)
	    {
		my $MSA_Pos1=$Reference_Pos_MSA_Pos{$Pos1};
		my $MSA_Pos2=$Reference_Pos_MSA_Pos{$Pos2};
		for my $Protein (keys %MSA)
		{
		    if ($Protein ne $Reference_Seq_ID)
		    {
			if ((@{$MSA{$Protein}}[$MSA_Pos2-1] ne "X") and (@{$MSA{$Protein}}[$MSA_Pos1-1] ne "X") and (@{$MSA{$Protein}}[$MSA_Pos1-1] ne "-") and (@{$MSA{$Protein}}[$MSA_Pos2-1] ne "-"))
			{
			    # SUM UP OVERLAPS
			    $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}+1 if (defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'});
			    $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}=1 if (!defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'});
			}
		    }
		}
	    }
	}
    }
    
# CREATING CLUSTERS HEAD CONSENSUS CONTACTS MAP
    my $input = new Bio::TreeIO(-file   => $Tree,
				-format => "newick");
    my $tree = $input->next_tree;
    my $Clusters_Ref=Analyze_Tree_Clusters1_of_Predict_WMC($tree,$Target_Name); #Only Clusters of Redundant
    
    
    my %Clusters_Contact_Maps;
    my %Clusters=%$Clusters_Ref;
    print "OUTSIDE ANALIZE CLUSTER FUNCTION\n########################################\n";
    foreach my $Head_Cluster_Node (keys %Clusters)
    {
	print "\nHEAD:",$Clusters{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id"),"\n";
	foreach my $Cluster_Member (@{$Clusters{"$Head_Cluster_Node"}{'Sons'}})
	{
	    if (defined $Cluster_Member)
	    {
		print "\tMEMBER:\t",$Cluster_Member->id,"\t";
		print $tree->distance(-nodes => [$Cluster_Member,$Clusters{"$Head_Cluster_Node"}{'Node'}]),"\n";
	    }
	}
	print "\n";
    }
    for (my $Pos1=1;$Pos1<$Reference_Seq_Size;$Pos1++)
    {
	for (my $Pos2=$Pos1+$MINIMAL_SEPARATION;$Pos2<$Reference_Seq_Size;$Pos2++)
	{
	    my $MSA_Pos1=$Reference_Pos_MSA_Pos{$Pos1};
	    my $MSA_Pos2=$Reference_Pos_MSA_Pos{$Pos2};
	    for my $Cluster_Head_ID (keys %Clusters)
	    {
		for my $Cluster_Member (@{$Clusters{"$Cluster_Head_ID"}{'Sons'}})
		{
		    if (defined $Cluster_Member)
		    {
			my $Cluster_Member_Name=$Cluster_Member->id;
			if ($Cluster_Member_Name=~/([A-Za-z0-9]{4}_[0-9A-Za-z])_ATOMS/)
			{
			    $Cluster_Member_Name=$1;
			}
			elsif ($Cluster_Member_Name=~/([A-Za-z0-9]{4}_[0-9A-Za-z]) ATOMS/)
			{
			    $Cluster_Member_Name=$1;
			}
			my $Num_Of_Cluster_Members=@{$Clusters{"$Cluster_Head_ID"}{'Sons'}};
			if ($Cluster_Member_Name ne $Target_Name)
			{
			    my $Weight;
			    my $Distance_To_Cluster_Head=$tree->distance(-nodes => [$Cluster_Member,$Clusters{"$Cluster_Head_ID"}{'Node'}]);
                            # ALL THE MEMBER OF THE CLUSTER HAS THE SAME WEIGHT
			    $Weight=1;
			    if (defined $MSA_Pos_to_Without_Gap_Pos{$Cluster_Member_Name}{$MSA_Pos1-1} and defined $MSA_Pos_to_Without_Gap_Pos{$Cluster_Member_Name}{$MSA_Pos2-1})
			    {
				if (defined $Contact_Maps_Hash{$Cluster_Member_Name}{$MSA_Pos_to_Without_Gap_Pos{$Cluster_Member_Name}{$MSA_Pos1-1}}{$MSA_Pos_to_Without_Gap_Pos{$Cluster_Member_Name}{$MSA_Pos2-1}})
				{
				    if (defined $Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2})
				    {
					if (defined $Contact_Maps_Hash{$Cluster_Member_Name}{$MSA_Pos_to_Without_Gap_Pos{$Cluster_Member_Name}{$MSA_Pos1-1}}{$MSA_Pos_to_Without_Gap_Pos{$Cluster_Member_Name}{$MSA_Pos2-1}})
					{
					    $Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}=$Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}+$Weight*1;
					    $Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Overlaps'}=$Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Overlaps'}+1;
					}
				    }
				    else
				    {
					if (defined $Contact_Maps_Hash{$Cluster_Member_Name}{$MSA_Pos_to_Without_Gap_Pos{$Cluster_Member_Name}{$MSA_Pos1-1}}{$MSA_Pos_to_Without_Gap_Pos{$Cluster_Member_Name}{$MSA_Pos2-1}})
					{
					    $Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}=$Weight*1;
					    $Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Overlaps'}=1;
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	}
    }
    print "FINISH CLUSTERING\n";#<STDIN>;
# FOR CLUSTERS HEADS
    my $root = $tree->get_root_node;
    my $Protein_Of_Interest_Node;
    my @leaves = $tree->get_leaf_nodes;
    my @Nodes= $tree->get_nodes();
    my @Internal_Nodes;
    my %Tree_Nodes;
    
# FIND PROTEIN OF INTEREST NODE
    foreach my $leaf1 (@leaves) 
    {
	print $leaf1->id,"\n";
	if ($leaf1->id eq $Target_Name)
	{
	    $Protein_Of_Interest_Node=$leaf1; 
	}
    }
    for (my $Pos1=1;$Pos1<$Reference_Seq_Size;$Pos1++)
    {
	for (my $Pos2=$Pos1+$MINIMAL_SEPARATION;$Pos2<$Reference_Seq_Size;$Pos2++)
	{
	    my $MSA_Pos1=$Reference_Pos_MSA_Pos{$Pos1};
	    my $MSA_Pos2=$Reference_Pos_MSA_Pos{$Pos2};
	    for my $Cluster_Head_ID (keys %Clusters_Contact_Maps)
	    {
		if (!$Entropy_Like) # OVERLAPS WERE NOT CALCULATED YET
		{
		    # SUM UP OVERLAPS
		    $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}+1 if (defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'});
		    $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}=1 if (!defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'});
		}
		#SUM UP THE WHIGHTS
		my $Distance_To_Protein_Of_Interest=$tree->distance(-nodes => [$Clusters{"$Cluster_Head_ID"}{'Node'},$Protein_Of_Interest_Node]);
		
		
		##### Taking Care For Distance To Protein Of Interest
		if ($Distance_To_Protein_Of_Interest<0.01) {$Distance_To_Protein_Of_Interest=0.01;}
		$Distance_To_Protein_Of_Interest=$Distance_To_Protein_Of_Interest if ($Power==1);
		$Distance_To_Protein_Of_Interest=$Distance_To_Protein_Of_Interest**$Power if ($Power>1);
		
		
		my $Local_Weight=1/$Distance_To_Protein_Of_Interest;
		$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'}+$Local_Weight if (defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'} and (defined $PhyloGenetic_Mode or $PhyloGenetic_Sum_Mode));
		$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'}=$Local_Weight if (!defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'} and  ($PhyloGenetic_Mode or $PhyloGenetic_Sum_Mode));
		
		# CHEAK FOR INTERACT
		if (defined $Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}) # INTERACTES
		{
		    print "INTERACT\n" if ($Debug_Mode);
		    if (defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'})
		    {
			$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}+($Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}/$Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Overlaps'}) if (!$PhyloGenetic_Mode and !$PhyloGenetic_Sum_Mode);
			$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}+(($Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}/$Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Overlaps'})*$Local_Weight) if ($PhyloGenetic_Mode or $PhyloGenetic_Sum_Mode);
		    }
		    else
		    {
			$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=($Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}/$Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Overlaps'});
			$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=(($Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}/$Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Overlaps'})*$Local_Weight);
		    }
		}
		elsif (!defined $Clusters_Contact_Maps{$Cluster_Head_ID}{$MSA_Pos1}{$MSA_Pos2}{'Agrees'}) # DON'T INTERACTS
		{
		    if (!defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'})
		    {
			$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=0;
		    }
		    print "DONT_INTERACT\n" if ($Debug_Mode);
		}
	    }
	    
	}
    }
    
    
#PRINT PREDICTED CONTACT MAP
    $Num_Of_Homologs=@Homologs_Structures;
    open (HOMOLOGS_CONSENSUS_CONTACT_MAP,">$Consensus_ContactMap") || die "Can't Open Consensus Contact Map: '$Consensus_ContactMap' $!";
    print HOMOLOGS_CONSENSUS_CONTACT_MAP "Pos1\tPos2\tAA1\tAA2\tPrediction\n";
    for my $Pos1 (sort keys (%Consensus_Contact_Map))
    {
	for my $Pos2 (sort keys (%{$Consensus_Contact_Map{$Pos1}}))
	{
	    # NOT PHYLOGENTIC MODE - NO Weighted Average
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}/$Num_Of_Homologs)),"\n" if (!$Only_Overlap_Mode and !$PhyloGenetic_Mode and !$PhyloGenetic_Sum_Mode);
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}/$Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'})),"\n" if ($Only_Overlap_Mode and !$PhyloGenetic_Mode and !$PhyloGenetic_Sum_Mode);
	    
	    # PHYLOGENETIC MODE - WEIGHTED AVERAGE
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}/$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'})),"\n" if ($PhyloGenetic_Mode and !$Entropy_Like and !$PhyloGenetic_Sum_Mode); # DO NOT CONSIDER THE AMOUNT OF INFORMATION I.E NON GAPS 
	    
	    # PHYLOGENETIC SUM MODE - WHIGHTED_SUM (NOT AVERAGES)
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'})),"\n" if ($PhyloGenetic_Sum_Mode and !$Entropy_Like and !$PhyloGenetic_Mode); # DO NOT CONSIDER THE AMOUNT OF INFORMATION I.E NON GAPS 

	    # PHYLOGENETIC MODE - WEIGHTED AVERAGE AND CONSIDER THE AMOUNT OF INFORMATION I.E NON GAPS 
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}/$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'})),"\n" if ($PhyloGenetic_Mode and $Entropy_Like and (($Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}/$Num_Of_Homologs))>0.5 and !$PhyloGenetic_Sum_Mode); #  CONSIDER THE AMOUNT OF INFORMATION I.E The NUM OF NON GAPS 
	}
    }
}

####  Functions section  ####################################################
sub read_Contact_Map_To_Hash_of_Predict_WMC
  {

#####        CONTACT MAP FORMAT     ###########
###############################################
#Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance Residue_A_Name Residue_B_Name
#70 18 0 0 1.1 6.1 T E
#135 111 0 0 13.0 4.1 L I
    
    my $ContactMap_Pos1_Col=0;
    my $ContactMap_Pos1_Chain_Col=2;
    my $ContactMap_Pos1_AA_Col=6;
    my $ContactMap_Pos2_Col=1;
    my $ContactMap_Pos2_Chain_Col=3;
    my $ContactMap_Pos2_AA_Col=7;
    my $Contact_Map_Distance_Col=5;
    my $ContactMap_File=$_[0];
    my $ContactsMap_Hash_Ref=$_[1];
    
    my %ContactsMap_Hash=%$ContactsMap_Hash_Ref;
    my $File_Without_Path;
    my $PDB_ID;
    if ($ContactMap_File=~/([A-Za-z0-9._]+)$/)
      {
	$File_Without_Path=$1;
	if ($File_Without_Path=~/(.*).CSU(_Contact_Map)?/)
	  {
	    $PDB_ID=$1;
	  }
      }
    print "Reading PDB: $PDB_ID Contact Map...\n";
    open (CONTACT_MAP,$ContactMap_File) || die "Can't Open ContactMap File: '$ContactMap_File' $!";
    my $line=<CONTACT_MAP>; #Header
    $line=<CONTACT_MAP>; #Header
    $line=<CONTACT_MAP>; #Header
    #Intra Chain Interactions at: /dir/SBTP/itamarb/test_set/contacts4/1a3a.ent
    #
    #Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance
    #16 83 A A 16.5 2.9
    #16 83 B B 15.6 2.9
    
    while ($line=<CONTACT_MAP>)
      {
	my @ContactMap_line_array=split (/\s/,$line);
	my $Pos1=$ContactMap_line_array[$ContactMap_Pos1_Col];
	my $Pos2=$ContactMap_line_array[$ContactMap_Pos2_Col];
#	print "PDB:$PDB_ID,POS1:$Pos1,Pos2:$Pos2;\n";<STDIN>;#QA
	$ContactsMap_Hash{$PDB_ID}{$Pos1}{$Pos2}=[@ContactMap_line_array];
      }
    close (CONTACT_MAP);
    
    return \%ContactsMap_Hash;
  }

sub Create_MSA_Pos_to_Without_Gap_Pos_of_Predict_WMC
  {
    my $MSA=$_[0];
    my %MSA=%$MSA;
    my %MSA_Pos_to_Without_Gap_Pos=();
    foreach my $Template_Name (keys %MSA)
      {
	my $Without_Gap_Pos=0;
	my @Seq=@{$MSA{$Template_Name}};
	my $Seq_Length=@Seq;
	for (my $i=0;$i<$Seq_Length;$i++)
	  {
	    if ($Seq[$i] ne "-")
	      {
		 $Without_Gap_Pos++;
		 $MSA_Pos_to_Without_Gap_Pos{$Template_Name}{$i}=($Without_Gap_Pos);
	       }
	  }
      }
    return \%MSA_Pos_to_Without_Gap_Pos;
  }

sub Build_Refference_Seq_MSA_Pos_Hash_of_Predict_WMC
  {
    my $Refference_Seq_ID=$_[0];
    my $MSA_Hash_Ref=$_[1];
    my %MSA_Hash=%$MSA_Hash_Ref;
    my %Refference_Seq_MSA_Position;
    my @Refference_Seq=@{$MSA_Hash{$Refference_Seq_ID}};
    my $MSA_Leangth=@Refference_Seq;
    my $Refference_Seq_Pos=0;
    for (my $i=0;$i<$MSA_Leangth;$i++)
      {
	if ($Refference_Seq[$i] ne "-")
	  {
	    $Refference_Seq_Pos++;
	    $Refference_Seq_MSA_Position{$Refference_Seq_Pos}=($i+1);
	  }
      }
    return \%Refference_Seq_MSA_Position;
  }
     
sub read_MSA_To_Hash_of_Predict_WMC
  {
    my $MSA_File=$_[0];
    my %MSA=();
    open (MSA,$MSA_File) || die "Can't Ope the MSA File $MSA_File $!";
    my $Seq="";
    my $Header="";
    my $First_Loop=0;
    my $Alignment_Length=0;
    my $Refference_Seq_ID="";
    while (my $line=<MSA>) # Reading MSA File
      {
	if ($line=~/^\>/) #Header
	  {
	    if ($First_Loop>0)
	      {
		$MSA{$Header}=[split(//,$Seq)];
		$Alignment_Length=@{$MSA{$Header}};
		#		print "LENGTH: $Alignment_Length\n";
		if ($Refference_Seq_ID eq "") #THE FIRST SEQ IS THE REFERENCE SEQ
		  {
		    $Refference_Seq_ID=$Header;
		  }
		$Seq="";
		chomp ($line);
		$Header=$line;
		if ($Header=~/([A-Za-z0-9]{4}_[0-9A-Za-z]) ATOMS/)
		{
		    $Header=$1;
		}
		elsif ($Header=~/([A-Za-z0-9]{4}_[0-9A-Za-z])_ATOMS/)
		{
		    $Header=$1;
		}
#		print $Header;<STDIN>;
	      }
	    if ($First_Loop==0)
	      {
		$First_Loop=1;
		chomp ($line);
		$Header=$line;
	      }
	  }
	elsif ($line =~ /^[A-Z\-\*]+$/) #Seq Of United MSA
	  {
	    chomp ($line);
	    $Seq.=$line;
	  }
      }
    $MSA{$Header}=[split(//,$Seq)]; 
    return \%MSA; 
  }
  
sub Analyze_Tree_Clusters1_of_Predict_WMC # Clusters of close only (Only Redundant)
  {
    my $tree=$_[0];
#    my $Tree_File=$_[0];
    my $Protein_Of_Interest=$_[1];

#    my $input = new Bio::TreeIO(-file   => $Tree_File,
#				-format => "newick");
#    my $tree = $input->next_tree;

    my $root = $tree->get_root_node;
    my $Protein_Of_Interest_Node;
    my @leaves = $tree->get_leaf_nodes;
    my @Nodes= $tree->get_nodes();
    my @Internal_Nodes;
    my %Tree_Nodes;
    
    # FIND PROTEIN OF INTEREST NODE
    foreach my $leaf1 (@leaves) 
      {
	print $leaf1->id,"\n";
	if ($leaf1->id eq $Protein_Of_Interest)
	  {
	    $Protein_Of_Interest_Node=$leaf1; 
	  }
      }
    # START CLUSTERING
    my $Cluster_Counter=0;
    my $Cluster_Id="A";
    my %Clusters=();
    my %Dist_Clusters;
    foreach my $leaf (@leaves)
      {
	my $Father=$leaf->ancestor;
	my $Dist_to_ProtOfInterest=$tree->distance(-nodes => [$leaf,$Protein_Of_Interest_Node]);
	if (!defined $Dist_Clusters{$Dist_to_ProtOfInterest})
	  {
	    $Dist_Clusters{$Dist_to_ProtOfInterest}="$Father";
	    $Father->add_tag_value("id",$Cluster_Id);
	    $Cluster_Id++;
	    $Clusters{"$Father"}{'Node'}=$Father;
	  }
	else
	  {
	    my $Father_Id=$Dist_Clusters{$Dist_to_ProtOfInterest};
	    $Father=$Clusters{"$Father_Id"}{'Node'};
	    $Clusters{"$Father"}{'Node'}=$Father;
	  }
	
	push (@{$Clusters{"$Father"}{'Sons'}},$leaf);
	#    print "FATHER:",$Father->get_tag_values("id"),"\tSON: ",$leaf->id,"\n";
	
      }
    # PRINT FIRST CLUSTERS (BY DIST TO PROTEIN OF INTEREST)
    #    foreach my $Head_Cluster_Node (keys %Clusters)
    #      {
    #	print "\n",$Clusters{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id"),"\t",$tree->distance(-nodes => [$Clusters{"$Head_Cluster_Node"}{'Node'},$Protein_Of_Interest_Node]),"\n";
    #	foreach my $Cluster_Member (@{$Clusters{"$Head_Cluster_Node"}{'Sons'}})
    #	  {
    #	    print $Cluster_Member->id,"\t";
    #	    print $tree->distance(-nodes => [$Cluster_Member,$Clusters{"$Head_Cluster_Node"}{'Node'}]),"\t",$tree->distance(-nodes => [$Cluster_Member,$Protein_Of_Interest_Node]),"\n";
    #	  }
    #	print "\n\n\n"
    #      }

    ### UNITE CLUSTERS HEAD (IF THEY ARE TOO CLOSE TO EACH OTHER)
    #########################################################################
    my %United_Cluster;
    my %Clusters_Head_Dist;
    my %Joind_Clusters_Names;
    my %Cluster_Heads_Names;
    my %Added_Clusters;
    my %United_Names;
    foreach my $Head_Cluster_Node (sort keys %Clusters) 
      {
	foreach my $Head_Key (sort keys %Clusters)
	  {
	    if ((abs ($tree->distance(-nodes => [$Clusters{"$Head_Cluster_Node"}{'Node'},$Clusters{"$Head_Key"}{'Node'}]))<0.02) and ($Head_Key ne $Head_Cluster_Node)) # The Dist Between Clusters Head
	      {
		#QA PRINT
		#	print "====", $Clusters{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id")," and ",$Clusters{"$Head_Key"}{'Node'}->get_tag_values("id")," Can Be United ====\n";
		#		print $Clusters{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id"),":\n";
		#		foreach my $Cluster_Member (@{$Clusters{"$Head_Cluster_Node"}{'Sons'}})
		#		  {
		#		    print $Cluster_Member->id,"\t";
		#		    print $tree->distance(-nodes => [$Cluster_Member,$Clusters{"$Head_Cluster_Node"}{'Node'}]),"\t",$tree->distance(-nodes => [$Cluster_Member,$Protein_Of_Interest_Node]),"\n";
		#		  }
		#		print $Clusters{"$Head_Key"}{'Node'}->get_tag_values("id"),":\n";
		#		foreach my $Cluster_Member (@{$Clusters{"$Head_Key"}{'Sons'}})
		#		  {
		#		    print $Cluster_Member->id,"\t";
		#		    print $tree->distance(-nodes => [$Cluster_Member,$Clusters{"$Head_Key"}{'Node'}]),"\t",$tree->distance(-nodes => [$Cluster_Member,$Protein_Of_Interest_Node]),"\n";
		#		  }
		my @Sons;
		# print $Clusters{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id"),"\n"; #QA
		my $Head_Cluster_id="$Head_Cluster_Node";
		my $Added_Cluster_id="$Head_Key";
		#print "check: $Added_Cluster_id and $Head_Cluster_id\n"; #QA
		if (!defined $Joind_Clusters_Names{$Head_Cluster_id}{$Added_Cluster_id} and !defined $Joind_Clusters_Names{$Added_Cluster_id}{$Head_Cluster_id} and !defined $Added_Clusters{$Added_Cluster_id} and !defined $Added_Clusters{$Head_Cluster_id}) # The Pair Was Not United already and The added cluster not were added elsewere
		  {
		    $United_Cluster{"$Head_Cluster_Node"}{'Node'}=$Clusters{"$Head_Cluster_Node"}{'Node'};
		    if (defined $United_Cluster{"$Head_Cluster_Node"}{'Sons'})
		      {
			@Sons=(@{$Clusters{"$Head_Key"}{'Sons'}},@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}});
		      }
		    else
		      {
			@Sons=(@{$Clusters{"$Head_Key"}{'Sons'}},@{$Clusters{"$Head_Cluster_Node"}{'Sons'}});
		      }
		    $United_Cluster{"$Head_Cluster_Node"}{'Sons'}=[@Sons];
		    $Joind_Clusters_Names{$Added_Cluster_id}{$Head_Cluster_id}=1; 
		    $Added_Clusters{$Added_Cluster_id}=1;
		    $United_Names{$Added_Cluster_id}=1;
		    $United_Names{$Head_Cluster_id}=1;
		  }
		# QA PRINT
		#		else
		#		  {
		#		    print $Clusters{"$Added_Cluster_id"}{'Node'}->get_tag_values("id"), " and ",$Clusters{"$Head_Cluster_id"}{'Node'}->get_tag_values("id")," were Already United or ",$Clusters{"$Added_Cluster_id"}{'Node'}->get_tag_values("id"), " were Added to Another Cluster\n";
		#		  }
	      }
	  }
      }
     # FIND THE CLUSTERS THAT WERE NOT UNITED AND ADD THEM TO THE FINAL LIST
    foreach my $Head_Cluster_Node (sort keys %Clusters) 
      {
	if (!defined $United_Names{"$Head_Cluster_Node"})
	  {
	    $United_Cluster{"$Head_Cluster_Node"}{'Sons'}=[@{$Clusters{"$Head_Cluster_Node"}{'Sons'}}];
	    $United_Cluster{"$Head_Cluster_Node"}{'Node'}=$Clusters{"$Head_Cluster_Node"}{'Node'};
	  }
      }
    ## QA PRINT UNITED CLUSTERS HEAD ##
    
    foreach my $Head_Cluster_Node (keys %United_Cluster)
      {
	print "\n",$United_Cluster{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id"),"\t",$tree->distance(-nodes => [$United_Cluster{"$Head_Cluster_Node"}{'Node'},$Protein_Of_Interest_Node]),"\n";
        foreach my $Cluster_Member (@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}})
	  {
	    print $Cluster_Member->id,"\t";
	    print $tree->distance(-nodes => [$Cluster_Member,$United_Cluster{"$Head_Cluster_Node"}{'Node'}]),"\t",$tree->distance(-nodes => [$Cluster_Member,$Protein_Of_Interest_Node]),"\n";
	  }
	print "\n"
      }
#    <STDIN>;
    # Split Cluster that their Seq are not redundant # DIFF From Analyze_Tree_Clusters
    foreach my $Head_Cluster_Node (sort keys %United_Cluster) 
      {
	my $Num_Of_Members=@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}};
	print $United_Cluster{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id"),"\t$Num_Of_Members\n";#<STDIN>;
	if ($Num_Of_Members==2) #Cluster of 2
	  {
	    print "   ### CLUSTER OF 2 ### \n";
	    my $Cluster_Member1=@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}}[0];
	    my $Cluster_Member2=@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}}[1];
	    my $Dist1_To_Head=$tree->distance(-nodes => [$Cluster_Member1,$United_Cluster{"$Head_Cluster_Node"}{'Node'}]);
	    my $Dist2_To_Head=$tree->distance(-nodes => [$Cluster_Member2,$United_Cluster{"$Head_Cluster_Node"}{'Node'}]);
	    if ($Dist1_To_Head>0.02 or $Dist2_To_Head>0.02)
	      {
		print "#### SPLIT ###\n";
		print $Cluster_Member1->id,"\t",$Dist1_To_Head,"\n";
		print $Cluster_Member2->id,"\t",$Dist2_To_Head,"\n";
		delete @{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}}[1];
		$United_Cluster{"$Head_Cluster_Node"}{'Node'}=undef;
		# Son1 Become Cluster Head and The Only Member
		$Cluster_Id++;
		$Cluster_Member1->add_tag_value("id",$Cluster_Id);
		$United_Cluster{"$Cluster_Member1"}{'Node'}=$Cluster_Member1; 
		$United_Cluster{"$Cluster_Member1"}{'Sons'}=[$Cluster_Member1];
		
		# Son2 Become Cluster Head and The Only Member
		$Cluster_Id++;
		$Cluster_Member2->add_tag_value("id",$Cluster_Id);
		$United_Cluster{"$Cluster_Member2"}{'Node'}=$Cluster_Member2; 
		$United_Cluster{"$Cluster_Member2"}{'Sons'}=[$Cluster_Member2];
		delete $United_Cluster{"$Head_Cluster_Node"}{'Node'};
		delete $United_Cluster{"$Head_Cluster_Node"};
#		print "delete United_Cluster{\"$Head_Cluster_Node\"}:$United_Cluster{\"$Head_Cluster_Node\"}\n"
	      }
	  }
	# DEALING WITH CLUSTERS OF MORE THAN 2
	elsif ($Num_Of_Members>2)
	  {
	    
	    for (my $i=0;$i<$Num_Of_Members;$i++)
	      {
		my $Cluster_Member=@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}}[$i];
		my $Dist_To_Head=$tree->distance(-nodes => [$Cluster_Member,$United_Cluster{"$Head_Cluster_Node"}{'Node'}]);
		if ($Dist_To_Head>0.05)
		  {
		    print "#### SPLIT ###\n";
		    print $Cluster_Member->id,"\t",$Dist_To_Head,"\n";
		    # Son Become Cluster Head and The Only Member
		    $Cluster_Id++;
		    $Cluster_Member->add_tag_value("id",$Cluster_Id);
		    $United_Cluster{"$Cluster_Member"}{'Node'}=$Cluster_Member; 
		    $United_Cluster{"$Cluster_Member"}{'Sons'}=[$Cluster_Member];
		    # Delete From The Old Cluster
		    delete @{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}}[$i];
		  }
	      }
	  }
	
      }
    

#    # FIND THE CLUSTERS THAT WERE NOT UNITED AND ADD THEM TO THE FINAL LIST
#    foreach my $Head_Cluster_Node (sort keys %Clusters) 
#      {
#	if (!defined $United_Names{"$Head_Cluster_Node"})
#	  {
#	    $United_Cluster{"$Head_Cluster_Node"}{'Sons'}=[@{$Clusters{"$Head_Cluster_Node"}{'Sons'}}];
#	    $United_Cluster{"$Head_Cluster_Node"}{'Node'}=$Clusters{"$Head_Cluster_Node"}{'Node'};
#	  }
#      }
    ## ALL The One Member Clusters Are Clusters Head
    foreach my $Head_Cluster_Node (sort keys %United_Cluster) 
      {
	my $Num_Of_Members=@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}};
	if ($Num_Of_Members==1)
	  {
	    my $Cluster_Member=@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}}[0];
	    #	    my $Cluster_Id=$United_Cluster{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id");
	    my $Cluster_Id=$Cluster_Member->id;
	    if (defined ($Cluster_Member->get_tag_values("id"))) {$Cluster_Member->remove_tag("id");}
	    $Cluster_Member->add_tag_value("id",$Cluster_Id);
	    $United_Cluster{"$Head_Cluster_Node"}{'Node'}=$Cluster_Member;
	  }
      }
    

    ## QA PRINT UNITED CLUSTERS HEAD ##
    
    foreach my $Head_Cluster_Node (keys %United_Cluster)
      {
	print "\n",$United_Cluster{"$Head_Cluster_Node"}{'Node'}->get_tag_values("id"),"\t",$tree->distance(-nodes => [$United_Cluster{"$Head_Cluster_Node"}{'Node'},$Protein_Of_Interest_Node]),"\n";
        foreach my $Cluster_Member (@{$United_Cluster{"$Head_Cluster_Node"}{'Sons'}})
	  {
	    if (defined $Cluster_Member)
	      {
		print $Cluster_Member->id,"\t";
		print $tree->distance(-nodes => [$Cluster_Member,$United_Cluster{"$Head_Cluster_Node"}{'Node'}]),"\t",$tree->distance(-nodes => [$Cluster_Member,$Protein_Of_Interest_Node]),"\n";
	      }
	  }
	print "\n"
      }
#    <STDIN>;
    return \%United_Cluster;
}


### END ###


## FROM: Templates_Consensus_Contacts_Predictor.pl
sub Predict_Consensus
{
    my ($ContactMap_List,$MSA,$Consensus_ContactMap,$Debug_Mode,$Only_Overlap_Mode,$Distance_Matrix,$Entropy_Like,$PhyloGenetic_Mode,$PhyloGenetic_Sum_Mode,$Power);
    $ContactMap_List=shift;
    $MSA=shift;
    $Consensus_ContactMap=shift;
    $Distance_Matrix=shift;
    $Power=shift;
    $Only_Overlap_Mode=shift;
    $Entropy_Like=shift;
    $PhyloGenetic_Mode=shift;
    $PhyloGenetic_Sum_Mode=shift;
    $Debug_Mode=shift;
#    PrmArgv ("Contact_Maps_List= %s ! File With Contacts Map List Of Templates", \$ContactMap_List,
#	     "MSA= %s ! Targets and Templates MSA", \$MSA,
#	     "Out= %s ! Consensus Contact Map According To the Homologs", \$Consensus_ContactMap,
#	     "Distances= %s ! Sequences Distance Matrix Compered to Target Protein - OPTIONAL for Weighted Average", \$Distance_Matrix,
#	     "-Debug ~ ! Flag For Printing Debug Info", \$Debug_Mode,
#	     "-Only_Overlap ~ ! Flag For considering only the overlap pairs in the averaging not all templates", \$Only_Overlap_Mode,
#	     "-Entropy_Like ~ ! Flag For considering the amount of information of the paired column", \$Entropy_Like,
#	     "-Phylogenetic_Mode ~ ! Flag For Consider Phylogenetic (Whighted Average) Information of the paired column", \$PhyloGenetic_Mode,
#	     "-Phylogenetic_Sum_Mode ~ ! Flag For Consider Phylogenetic (Whighted Sum) Information of the paired column", \$PhyloGenetic_Sum_Mode,
#	     "Power=1 %s ! Considering Power of distance", \$Power,
#	     );
    
    my $MINIMAL_SEPARATION=6;
#####        CONTACT MAP FORMAT     ###########
###############################################
#Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance Residue_A_Name Residue_B_Name
#70 18 0 0 1.1 6.1 T E
#135 111 0 0 13.0 4.1 L I

    my $ContactMap_Pos1_Col=0;
    my $ContactMap_Pos1_Chain_Col=2;
    my $ContactMap_Pos1_AA_Col=6;
    my $ContactMap_Pos2_Col=1;
    my $ContactMap_Pos2_Chain_Col=3;
    my $ContactMap_Pos2_AA_Col=7;
    my $Contact_Map_Distance_Col=5;
    
################################################
    my %DistanceMatrix=();
    
    my $Reference_Seq_ID;
    my %Consensus_Contact_Map=();
    open (MSA,$MSA) || die "Can't Open MSA: '$MSA' $!";
    $Reference_Seq_ID=<MSA>;
    chomp ($Reference_Seq_ID);
    close (MSA);
    my @Homologs_Structures;
    my $MSA_Hash_Ref=read_MSA_To_Hash_of_Predict_Consensus($MSA);
    my %MSA=%$MSA_Hash_Ref;
    print "=== Finished Reading MSA: $MSA To Hash\n";
    my $Tmp=join("",@{$MSA{$Reference_Seq_ID}});
    $Tmp =~ s/-//g;
    my $Reference_Seq=$Tmp;
my @Reference_Seq_Array=("START_FROM1",split(//,$Reference_Seq));
    print "Reference Seq: $Tmp\n";
    my $Reference_Pos_MSA_Pos_Ref=Build_Refference_Seq_MSA_Pos_Hash_of_Predict_Consensus ($Reference_Seq_ID,$MSA_Hash_Ref);
    my %Reference_Pos_MSA_Pos=%$Reference_Pos_MSA_Pos_Ref;
    my $MSA_Pos_to_Without_Gap_Pos=Create_MSA_Pos_to_Without_Gap_Pos_of_Predict_Consensus($MSA_Hash_Ref);
    my %MSA_Pos_to_Without_Gap_Pos=%$MSA_Pos_to_Without_Gap_Pos;
    
    my %Contact_Maps_Hash=();
    my $Contact_Maps_Hash_Ref;
    
    if (defined $Distance_Matrix)
    {
	my $Min_Distance=99999;
	open (DISTANCES,$Distance_Matrix) || die "Can't Open Distance Matrix File '$Distance_Matrix' $!";
	while (my $line=<DISTANCES>)
	{
	    my @line=split(/\t/,$line);
	    if ($line[1] ne $line[0])
	    {
		my $Template_PDB;
		if ($line[1]=~/([0-9A-Za-z]{4}_[A-Za-z0-9]) ATOMS/)
		{
		    $Template_PDB=$1;
		}
		elsif ($line[1]=~/([0-9A-Za-z]{4}_[A-Za-z0-9])_ATOMS/)
		{
		    $Template_PDB=$1;
		}
		my $Distance=$line[2];
		# Finding Min Distance
		if ($Distance>0.01)
		{
		    if ($Distance<$Min_Distance)
		    {
			$Min_Distance=$Distance;
		    }
		}
	    }
	}
	close (DISTANCES);
	open (DISTANCES,$Distance_Matrix) || die "Can't Open Distance Matrix File '$Distance_Matrix' $!";
	while (my $line=<DISTANCES>)
	{
	    my @line=split(/\t/,$line);
	    if ($line[1] ne $line[0])
	    {
		my $Template_PDB;
		if ($line[1]=~/([0-9A-Za-z]{4}_[A-Za-z0-9]) ATOMS/)
		{
		    $Template_PDB=$1;
		}
		elsif ($line[1]=~/([0-9A-Za-z]{4}_[A-Za-z0-9])_ATOMS/)
		{
		    $Template_PDB=$1;
		}
		my $Distance=$line[2];
		if ($Distance<0.01)
		{
		    $Distance=0.01;
		}
		$DistanceMatrix{$Reference_Seq_ID}{$Template_PDB}=$Distance if ($Power==1); # The Weight Will Be 1/Distance
		$DistanceMatrix{$Reference_Seq_ID}{$Template_PDB}=$Distance**$Power if ($Power>1); # The Weight Will Be 1/Distance^Power
	    }
	}
	close (DISTANCES);
    }
    open (CONATACTMAP_LIST,$ContactMap_List) || die "Can't Open Contact Map List: '$ContactMap_List' $!";
    while (my $ContactMap_File=<CONATACTMAP_LIST>)
    {
	my $File_Without_Path;
	my $PDB_ID;
	$Contact_Maps_Hash_Ref=read_Contact_Map_To_Hash_of_Predict_Consensus($ContactMap_File,\%Contact_Maps_Hash);
	%Contact_Maps_Hash=%$Contact_Maps_Hash_Ref;
	## Q.A
	if ($ContactMap_File=~/([A-Za-z0-9._]+)$/)
	{
	    $File_Without_Path=$1;
	    if ($File_Without_Path=~/(.*).CSU_Contact_Map/)
	    {
		$PDB_ID=$1;
	    }
	    push (@Homologs_Structures,$PDB_ID);
	}
    }
    %Contact_Maps_Hash=%$Contact_Maps_Hash_Ref;
    
    my $Num_Of_Homologs=@Homologs_Structures;
    my $Reference_Seq_Size=@Reference_Seq_Array;			 
    if ($Entropy_Like) # CONSIDER THE AMOUNT OF INFORMATION IN EACH PAIRED COLUMN
    {
	# COUNTING PAIRS
	for (my $Pos1=1;$Pos1<$Reference_Seq_Size;$Pos1++)
	{
	    for (my $Pos2=$Pos1+$MINIMAL_SEPARATION;$Pos2<$Reference_Seq_Size;$Pos2++)
	    {
		my $MSA_Pos1=$Reference_Pos_MSA_Pos{$Pos1};
		my $MSA_Pos2=$Reference_Pos_MSA_Pos{$Pos2};
		for my $Protein (keys %MSA)
		{
		    if ($Protein ne $Reference_Seq_ID)
		    {
			if ((@{$MSA{$Protein}}[$MSA_Pos2-1] ne "X") and (@{$MSA{$Protein}}[$MSA_Pos1-1] ne "X") and (@{$MSA{$Protein}}[$MSA_Pos1-1] ne "-") and (@{$MSA{$Protein}}[$MSA_Pos2-1] ne "-"))
			{
			    # SUM UP OVERLAPS
			    $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}+1 if (defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'});
			    $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}=1 if (!defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'});
			}
		    }
		}
	    }
	}
    }
    
    for (my $Pos1=1;$Pos1<$Reference_Seq_Size;$Pos1++)
    {
	for (my $Pos2=$Pos1+$MINIMAL_SEPARATION;$Pos2<$Reference_Seq_Size;$Pos2++)
	{
	    my $MSA_Pos1=$Reference_Pos_MSA_Pos{$Pos1};
	    my $MSA_Pos2=$Reference_Pos_MSA_Pos{$Pos2};
	    for my $Protein (keys %MSA)
	    {
		if ($Protein ne $Reference_Seq_ID)
		{
		    if ((@{$MSA{$Protein}}[$MSA_Pos2-1] ne "X") and (@{$MSA{$Protein}}[$MSA_Pos1-1] ne "X") and (@{$MSA{$Protein}}[$MSA_Pos1-1] ne "-") and (@{$MSA{$Protein}}[$MSA_Pos2-1] ne "-"))
		    {
			print "PDB: $Protein\tMSA_Pos1:",$MSA_Pos1,"\tMSA_AA1:@{$MSA{$Protein}}[$MSA_Pos1-1]\tMSA_Pos2:",$MSA_Pos2,"\tMSA_AA2:@{$MSA{$Protein}}[$MSA_Pos2-1]\tSeq_Pos1: $Reference_Seq_Array[$Pos1]\tSeq_Pos2: $Reference_Seq_Array[$Pos2]\tSeq_Pos1:",$MSA_Pos_to_Without_Gap_Pos{$Protein}{$MSA_Pos1-1},"\tSeq_Pos2:",$MSA_Pos_to_Without_Gap_Pos{$Protein}{$MSA_Pos2-1},"\t$Pos1\t$Pos2\t" if ($Debug_Mode);
			if (!$Entropy_Like) # OVERLAPS WERE NOT CALCULATED YET
			{
			    # SUM UP OVERLAPS
			    $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}+1 if (defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'});
			    $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}=1 if (!defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'});
			}
			#SUM UP THE WHIGHTS
			my $Local_Weight=1/$DistanceMatrix{$Reference_Seq_ID}{$Protein} if ($PhyloGenetic_Mode or $PhyloGenetic_Sum_Mode);
			#$Local_Weight=(1/$DistanceMatrix{$Reference_Seq_ID}{$Protein})*($Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}/$Num_Of_Homologs) if ($Entropy_Like);
			
			$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'}+$Local_Weight if (defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'} and (defined $PhyloGenetic_Mode or $PhyloGenetic_Sum_Mode));
			$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'}=$Local_Weight if (!defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'} and  ($PhyloGenetic_Mode or $PhyloGenetic_Sum_Mode));
			
			# CHEAK FOR INTERACT
			if (defined $Contact_Maps_Hash{$Protein}{$MSA_Pos_to_Without_Gap_Pos{$Protein}{$MSA_Pos1-1}}{$MSA_Pos_to_Without_Gap_Pos{$Protein}{$MSA_Pos2-1}}) # INTERACTES
			{
			    print "INTERACT\n" if ($Debug_Mode);
			    if (defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'})
			    {
				$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}+1 if (!$PhyloGenetic_Mode and !$PhyloGenetic_Sum_Mode);
				$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}+(1*$Local_Weight) if ($PhyloGenetic_Mode or $PhyloGenetic_Sum_Mode);
			    }
			    else
			    {
				$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=1;
				$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=(1*$Local_Weight) if ($PhyloGenetic_Mode or $PhyloGenetic_Sum_Mode); 
			    }
			}
			elsif (!defined $Contact_Maps_Hash{$Protein}{$MSA_Pos_to_Without_Gap_Pos{$Protein}{$MSA_Pos1-1}}{$MSA_Pos_to_Without_Gap_Pos{$Protein}{$MSA_Pos2-1}}) # DON'T INTERACTS
			{
			    if (!defined $Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'})
			    {
				$Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}=0;
			    }
			    print "DONT_INTERACT\n" if ($Debug_Mode);
			}
		    }
		    
		}
	    }
	}
    }
    
#PRINT CONSENSUS CONTACT MAP
    $Num_Of_Homologs=@Homologs_Structures;
    open (HOMOLOGS_CONSENSUS_CONTACT_MAP,">$Consensus_ContactMap") || die "Can't Open Consensus Contact Map: '$Consensus_ContactMap' $!";
    print HOMOLOGS_CONSENSUS_CONTACT_MAP "Pos1\tPos2\tAA1\tAA2\tPrediction\n";
    for my $Pos1 (sort keys (%Consensus_Contact_Map))
    {
	for my $Pos2 (sort keys (%{$Consensus_Contact_Map{$Pos1}}))
	{
	    # NOT PHYLOGENTIC MODE - NO Weighted Average
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}/$Num_Of_Homologs)),"\n" if (!$Only_Overlap_Mode and !$PhyloGenetic_Mode and !$PhyloGenetic_Sum_Mode);
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}/$Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'})),"\n" if ($Only_Overlap_Mode and !$PhyloGenetic_Mode and !$PhyloGenetic_Sum_Mode);
	    
	    # PHYLOGENETIC MODE - WEIGHTED AVERAGE
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}/$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'})),"\n" if ($PhyloGenetic_Mode and !$Entropy_Like and !$PhyloGenetic_Sum_Mode); # DO NOT CONSIDER THE AMOUNT OF INFORMATION I.E NON GAPS 
	    
	    # PHYLOGENETIC SUM MODE - WHIGHTED_SUM (NOT AVERAGES)
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'})),"\n" if ($PhyloGenetic_Sum_Mode and !$Entropy_Like and !$PhyloGenetic_Mode); # DO NOT CONSIDER THE AMOUNT OF INFORMATION I.E NON GAPS 
	    
	    # PHYLOGENETIC MODE - WEIGHTED AVERAGE AND CONSIDER THE AMOUNT OF INFORMATION I.E NON GAPS 
	    print HOMOLOGS_CONSENSUS_CONTACT_MAP "$Pos1\t$Pos2\t$Reference_Seq_Array[$Pos1]\t$Reference_Seq_Array[$Pos2]\t",sprintf("%.4f",($Consensus_Contact_Map{$Pos1}{$Pos2}{'Agrees'}/$Consensus_Contact_Map{$Pos1}{$Pos2}{'Weights'})),"\n" if ($PhyloGenetic_Mode and $Entropy_Like and (($Consensus_Contact_Map{$Pos1}{$Pos2}{'Overlapped'}/$Num_Of_Homologs))>0.5 and !$PhyloGenetic_Sum_Mode); #  CONSIDER THE AMOUNT OF INFORMATION I.E The NUM OF NON GAPS 
	    
	}
    }
}
####  Functions section  ####################################################
sub read_Contact_Map_To_Hash_of_Predict_Consensus
  {
    my $ContactMap_File=$_[0];
    my $ContactsMap_Hash_Ref=$_[1];
    
    my %ContactsMap_Hash=%$ContactsMap_Hash_Ref;
    my $File_Without_Path;
    my $PDB_ID;
    if ($ContactMap_File=~/([A-Za-z0-9._]+)$/)
      {
	$File_Without_Path=$1;
	if ($File_Without_Path=~/(.*).CSU(_Contact_Map)?/)
	  {
	    $PDB_ID=$1;
	  }
      }
    print "Reading PDB: $PDB_ID Contact Map...\n";
    open (CONTACT_MAP,$ContactMap_File) || die "Can't Open ContactMap File: '$ContactMap_File' $!";
    my $line=<CONTACT_MAP>; #Header
    $line=<CONTACT_MAP>; #Header
    $line=<CONTACT_MAP>; #Header
    #Intra Chain Interactions at: /dir/SBTP/itamarb/test_set/contacts4/1a3a.ent
    #
    #Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance
    #16 83 A A 16.5 2.9
    #16 83 B B 15.6 2.9
    my $ContactMap_Pos1_Col=0;
    my $ContactMap_Pos1_Chain_Col=2;
    my $ContactMap_Pos1_AA_Col=6;
    my $ContactMap_Pos2_Col=1;
    my $ContactMap_Pos2_Chain_Col=3;
    my $ContactMap_Pos2_AA_Col=7;
    my $Contact_Map_Distance_Col=5;
    while ($line=<CONTACT_MAP>)
      {
	my @ContactMap_line_array=split (/\s/,$line);
	my $Pos1=$ContactMap_line_array[$ContactMap_Pos1_Col];
	my $Pos2=$ContactMap_line_array[$ContactMap_Pos2_Col];
#	print "PDB:$PDB_ID,POS1:$Pos1,Pos2:$Pos2;\n";<STDIN>;#QA
	$ContactsMap_Hash{$PDB_ID}{$Pos1}{$Pos2}=[@ContactMap_line_array];
      }
    close (CONTACT_MAP);
    
    return \%ContactsMap_Hash;
  }

sub Create_MSA_Pos_to_Without_Gap_Pos_of_Predict_Consensus
  {
    my $MSA=$_[0];
    my %MSA=%$MSA;
    my %MSA_Pos_to_Without_Gap_Pos=();
    foreach my $Template_Name (keys %MSA)
      {
	my $Without_Gap_Pos=0;
	my @Seq=@{$MSA{$Template_Name}};
	my $Seq_Length=@Seq;
	for (my $i=0;$i<$Seq_Length;$i++)
	  {
	    if ($Seq[$i] ne "-")
	      {
		 $Without_Gap_Pos++;
		 $MSA_Pos_to_Without_Gap_Pos{$Template_Name}{$i}=($Without_Gap_Pos);
	       }
	  }
      }
    return \%MSA_Pos_to_Without_Gap_Pos;
  }

sub Build_Refference_Seq_MSA_Pos_Hash_of_Predict_Consensus
  {
    my $Refference_Seq_ID=$_[0];
    my $MSA_Hash_Ref=$_[1];
    my %MSA_Hash=%$MSA_Hash_Ref;
    my %Refference_Seq_MSA_Position;
    my @Refference_Seq=@{$MSA_Hash{$Refference_Seq_ID}};
    my $MSA_Leangth=@Refference_Seq;
    my $Refference_Seq_Pos=0;
    for (my $i=0;$i<$MSA_Leangth;$i++)
      {
	if ($Refference_Seq[$i] ne "-")
	  {
	    $Refference_Seq_Pos++;
	    $Refference_Seq_MSA_Position{$Refference_Seq_Pos}=($i+1);
	  }
      }
    return \%Refference_Seq_MSA_Position;
  }

sub read_MSA_To_Hash_of_Predict_Consensus
  {
    my $MSA_File=$_[0];
    my %MSA=();
    open (MSA,$MSA_File) || die "Can't Ope the MSA File $MSA_File $!";
    my $Seq="";
    my $Header="";
    my $First_Loop=0;
    my $Alignment_Length=0;
    my $Refference_Seq_ID="";
    while (my $line=<MSA>) # Reading MSA File
      {
	if ($line=~/^\>/) #Header
	  {
	    if ($First_Loop>0)
	      {
		$MSA{$Header}=[split(//,$Seq)];
		$Alignment_Length=@{$MSA{$Header}};
		#		print "LENGTH: $Alignment_Length\n";
		if ($Refference_Seq_ID eq "") #THE FIRST SEQ IS THE REFERENCE SEQ
		  {
		    $Refference_Seq_ID=$Header;
		  }
		$Seq="";
		chomp ($line);
		$Header=$line;
		if ($Header=~/([A-Za-z0-9]{4}_[0-9A-Za-z]) ATOMS/)
		{
		      $Header=$1;
		}
		elsif ($Header=~/([A-Za-z0-9]{4}_[0-9A-Za-z])_ATOMS/)
		{
		      $Header=$1;
		}
#		print $Header;<STDIN>;
	      }
	    if ($First_Loop==0)
	      {
		$First_Loop=1;
		chomp ($line);
		$Header=$line;
	      }
	  }
	elsif ($line =~ /^[A-Z\-\*]+$/) #Seq Of United MSA
	  {
	    chomp ($line);
	    $Seq.=$line;
	  }
      }
    $MSA{$Header}=[split(//,$Seq)]; 
    return \%MSA; 
  }


##  Add_CSU_Contact_Map_To_Prediction_File.pl
sub ADD_CSU_CONTACT_MAP_TO_PREDICTION_FILE
{
    my ($Prediction_File,$Out_File,$Start,$End,$Separation,$Reff_Seq_File,$PDB_ID,$Protein_ContactMap,$Tmp_Path);
    $Start=1;
    $Prediction_File=shift;
    $Out_File=shift;
    $Tmp_Path=shift;
    $Reff_Seq_File=shift;
    $PDB_ID=shift;
#    PrmArgv ("Prediction= %s ! The File Containing The Prediction", \$Prediction_File,
#	 "Out= %s ! Output File", \$Out_File,
#	 "Tmp_Path= %s ! Path For Tmp Files", \$Tmp_Path,
#	 "Start=1 %s ! Start Point", \$Start,
#	 "Reff_Seq= %s ! File With The Refference Seq (Simple Fasta Format)", \$Reff_Seq_File,
#	 "PDB_ID= %s ! The PDB ID", \$PDB_ID,
#	 "Contact_Map= %s ! Contact Map File to get (Optional)", \$Protein_ContactMap,
#	 "End= %s ! End Point (Optional, Default Protein Length)", \$End
#	);

    my %AA_Encoding=(G=>'1',A=>'2',V=>'3',L=>'4',I=>'5',M=>'6',F=>'7',W=>'8', P=>'9',S=>'10',T=>'11',C=>'12',Y=>'13',N=>'14',Q=>'15',D=>'16',E=>'17',K=>'18',R=>'19',H=>'20');

    open (OUT,">$Out_File") || die "Can't Open The Output File: $Out_File $!";
#PDB Location
#=============
    my $PDB_Path=WMC_CONSTANTS::PDB_LOCAL_PATH;

#Contact Map Script
#=====================

#####        CONTACT MAP FORMAT     ###########
###############################################
#Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance Residue_A_Name Residue_B_Name
#70 18 0 0 1.1 6.1 T E
#135 111 0 0 13.0 4.1 L I

    my $ContactMap_Pos1_Col=0;
    my $ContactMap_Pos1_Chain_Col=2;
    my $ContactMap_Pos1_AA_Col=6;
    my $ContactMap_Pos2_Col=1;
    my $ContactMap_Pos2_Chain_Col=3;
    my $ContactMap_Pos2_AA_Col=7;
    my $Contact_Map_Distance_Col=5;

#### PREDICTION FILE FORMAT   #######
#####################################
# HAIM FORMAT
    my $Prediction_File_Pos1_Col=0;
    my $Prediction_File_Pos2_Col=1;
    my $Prediction_File_Pos1_AA_Col=2;
    my $Prediction_File_Pos2_AA_Col=3;
    
# ITAMAR MI FORMAT
#my $Prediction_File_Pos1_Col=5;
#my $Prediction_File_Pos2_Col=9;
#my $Prediction_File_Pos1_AA_Col=6;
#my $Prediction_File_Pos2_AA_Col=10;
    
#Chek For PDB ID
#=====================
    my $PDB_ID_Flag=0;
    my $PDB_Exist=0;
    my $Tmp_PDB_File;
    my $Dont_Delete_Flag=0;
    if (defined $PDB_ID)
    {
	my $PDB_ID_Name=lc($PDB_ID);
	my $PDB_File=$PDB_Path.substr($PDB_ID_Name, 1,2)."/pdb".substr($PDB_ID_Name, 0,4).".ent.gz";
	$Tmp_PDB_File=$Tmp_Path.substr($PDB_ID_Name, 0,4).".ent";
	if (-e  $PDB_File)
	{
	    system ("zcat $PDB_File | grep -v \"ANISOU\" > $Tmp_PDB_File");
	    $PDB_Exist=1;
	}
	else 
	{
	    my $Downloaded_File=$MissingPDB_Path."pdb".substr($PDB_ID_Name, 0,4).".ent.gz";
	    my $WGET_PATH=WMC_CONSTANTS::PDB_WGET_PATH;
	    system("wget $WGET_PATH".substr($PDB_ID_Name, 1,2)."/pdb".substr($PDB_ID_Name, 0,4).".ent.gz -O $Downloaded_File");
	    if (-e $Downloaded_File)
	    {
		system ("zcat $Downloaded_File | grep -v \"ANISOU\" > $Tmp_PDB_File"); 
	    }
	    $PDB_Exist=1;
	}
    }
    if ($PDB_Exist!=1)
    {
	print "PDB: $PDB_ID WAS NOT FOUND\n";
	exit (1);
    }

#Extract The PDB Atoms Seq
##############################
    my $PDB_Chain_ID;
    my $PDB_Name;
    if ($PDB_ID =~ /^([A-Za-z0-9]{4})_([A-Za-z0-9]{1})/)
    {
	$PDB_Name=$1;
	$PDB_Chain_ID=$2;
    }
    my $Atoms_Seq_File=$Tmp_Path.$PDB_ID."_ATOMS.fs";
#print "EXTRACT ATOM SEQ\n";<STDIN>;
    my $Atoms_Seq=From_PDB_ATOMS_To_Seq_of_ADD_CSU_TO_PREDICTION_FILE($Tmp_PDB_File,$PDB_Chain_ID,$Atoms_Seq_File);
    
#Chek For Contact Map
#=====================
    my $Contact_Map_Flag=0;
    
# Cheack If Contact Map Exist
    if (defined $Protein_ContactMap) # The Protein Has Contact Map
    {
	if (-e $Protein_ContactMap) # File Exist
	{
	    $Contact_Map_Flag=1;
	}
    }
    elsif (defined $PDB_ID)
    {
	$Protein_ContactMap=$Tmp_Path."$PDB_ID".".CSU_Contact_Map";
	CSU_ContactMap_MMOL_Only_With_AA_Inter_Chain_Intra_Chain_Interactions1($Tmp_PDB_File,$Protein_ContactMap,$PDB_Chain_ID) if (!-e $Protein_ContactMap);
#	my $ContactMap_Command=$Contact_Map_Script." In=$Tmp_PDB_File Out= $Protein_ContactMap Chain_ID=$PDB_Chain_ID"; # HERE
#    print $ContactMap_Command;<STDIN>;
#	system ("$ContactMap_Command") if (!-e $Protein_ContactMap);
    }
    
#Find The Gap Between The Atoms Seq And Reference Seq
#######################################################
    my $Reffernce_Protein_Name;
    if ($Reff_Seq_File =~ /([A-Za-z0-9_.]+)$/)
    {
	$Reffernce_Protein_Name=$1;
    }
    my $Atoms_Reff_Seq_bl2seq="$Tmp_Path"."$PDB_ID"."_Atoms_"."$Reffernce_Protein_Name"."_bl2seq";
    my @OverLap_Start_End=gap_blast2seq_of_ADD_CSU_TO_PREDICTION_FILE($Atoms_Seq_File, $Reff_Seq_File, $Atoms_Reff_Seq_bl2seq);
    my $Refference_Seq_PDB_Atoms_Gap=$OverLap_Start_End[0];
    my $Overlap_Start=$OverLap_Start_End[1]; 
    my $Overlap_End=$OverLap_Start_End[2];
    my $Known_MisMatch=$OverLap_Start_End[4];
    my %Known_MisMatch=%$Known_MisMatch; # Mismatch between the Atoms And Referencer Seq
#print "GAP:$Refference_Seq_PDB_Atoms_Gap\n";<STDIN>;
    my $MISMATCH_FLAG=0;
    if ($Overlap_Start>$Start)
    {
	$Start=$Overlap_Start;
    }
    if (!defined $End)
    {
	$End=$Overlap_End;
    }
    if ($Overlap_End<$End) 
    {
	$End=$Overlap_End;
    }
    
# Create Hash To hold all the interactions
################################################
    my $Interactions_Hash_Ref=Builed_Interactions_Hash_CSU_of_ADD_CSU_TO_PREDICTION_FILE($Protein_ContactMap,$PDB_Chain_ID,$Refference_Seq_PDB_Atoms_Gap);
    my %Interactions_Hash=%$Interactions_Hash_Ref; ##Contains All Interactions
    unlink ($Tmp_PDB_File) if (!$Dont_Delete_Flag);
    
# Add The Interaction To Prediction File And Extract The Predictions Covers the PDB Segment
############################################################################################
    open (PREDICTION_FILE,$Prediction_File) || die "Can'r open Prediction File: $Prediction_File $!";
    my $line=<PREDICTION_FILE>;
    chomp ($line);
    print OUT "$line\tInteraction\n";
    my %Prediction_Hash;
    while (my $line=<PREDICTION_FILE>)
    {
	chomp ($line);
	my @line=split(/\t/,$line);
	my $Pos1=$line[$Prediction_File_Pos1_Col];
	my $Pos2=$line[$Prediction_File_Pos2_Col];
	my $AA1=$line[$Prediction_File_Pos1_AA_Col];
	my $AA2=$line[$Prediction_File_Pos2_AA_Col];
#    print "POS1:$Pos1\tPOS2:$Pos2\tAA1:$AA1\tAA2:$AA2\n";<STDIN>;
	if (($Pos1>=$Start)&&($Pos2<=$End)) # Covered By The PDB Segment
	{
	    if ((defined $Interactions_Hash{$Pos1}{$Pos2}) && ($Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos1_AA_Col] eq  $AA1) && ($Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos2_AA_Col] eq  $AA2))
	    {
#		print "AA1:$Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos1_AA_Col]\n";<STDIN>;
		print OUT "$line\t1\n";
	    }
	    elsif ((defined $Interactions_Hash{$Pos1}{$Pos2})&&(($Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos1_AA_Col] ne  $AA1) || ($Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos2_AA_Col] ne  $AA2)))
	    {
		if (($Known_MisMatch{$Pos1})||($Known_MisMatch{$Pos2}))
		{
		    print "###### KNOWN MISMATCH ERROR ON $PDB_ID: Interactions_Hash{$Pos1}{$Pos2}: @{$Interactions_Hash{$Pos1}{$Pos2}}\tInteractions_AA1:$Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos1_AA_Col]* Interactions_AA2:$Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos2_AA_Col]*\tAA1:$AA1* AA2:$AA2 #####\n";#<STDIN>;
		    }
		else
		{
		    print "###### UNKNOWN MISMATCH ERROR ON $PDB_ID: Interactions_Hash{$Pos1}{$Pos2}:  @{$Interactions_Hash{$Pos1}{$Pos2}}\tInteractions_AA1:$Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos1_AA_Col]* Interactions_AA2$Interactions_Hash{$Pos1}{$Pos2}[$ContactMap_Pos2_AA_Col]* AA1:$AA1* AA2:$AA2*#####\n";#<STDIN>;
		}    
	    }
	    elsif (!defined $Interactions_Hash{$Pos1}{$Pos2}) 
	    {
#	    print OUT "$line\t0.00000000001\n";
		print OUT "$line\t0\n";
	    }
	}
    }
}   
# FUNCTION SECTION
#####################
sub From_PDB_ATOMS_To_Seq_of_ADD_CSU_TO_PREDICTION_FILE
{
    # print "==================FROM PDB ATOMS TO SEQ======================\n";
    my $Protein_PDB_File=$_[0];
    my $Protein_Chain_ID=$_[1];
    my $Out_File=$_[2];
    # print "PDB: $Protein_PDB_File\n";<STDIN>;
    my %AA_Abb=(GLY=>'G',ALA=>'A',VAL=>'V',LEU=>'L',ILE=>'I',MET=>'M',PHE=>'F',TRP=>'W',PRO=>'P',SER=>'S',THR=>'T',CYS=>'C',TYR=>'Y',ASN=>'N',GLN=>'Q',ASP=>'D',GLU=>'E',LYS=>'K',ARG=>'R',HIS=>'H',);
    my $Protein_Name;
    if ($Protein_PDB_File=~/([1-9a-zA-Z.]+)$/)
    {
	$Protein_Name=$1;
    }
    my $first=0; #untile the first AA is found $first==0
    my $Prev_AA=0;
    my @Atoms_Seq;
    open (Protein_PDB,$Protein_PDB_File) || die "Can't open pdb file: $Protein_PDB_File $!\n";
    
    while (my $line=<Protein_PDB>)
    {
	my $AA_Name;
	if (($line=~/^ATOM/)&&($first==0))
	{
	    $first=1;
	    my $First_PDB_AA;
	    if ($line=~ /^.{21}([A-Z ]{1})([0-9 ]{4})/)
	    {
		$First_PDB_AA=$2;
		$First_PDB_AA=remove_whitespace($First_PDB_AA);
	    }
	}
	if ($line=~/^ATOM/)
	{
	    my $AA_Index;
	    my $PDB_Chain_ID;
	    if ($line=~ /^.{17}([A-Z]{3}).{1}([A-Z0-9 ]{1})([0-9 ]{4})/)
	    {
		$PDB_Chain_ID=$2;
		$AA_Index=$3;
		$AA_Name=$1;
		$AA_Name=remove_whitespace($AA_Name);
		$AA_Index=remove_whitespace($AA_Index);
		$PDB_Chain_ID=remove_whitespace($PDB_Chain_ID);
		if ($PDB_Chain_ID eq "")
		{
		    $PDB_Chain_ID=0;
		}
	    }
#	    print "PDB_Chain_ID:$PDB_Chain_ID\tProtein_Chain_ID:$Protein_Chain_ID\n";<STDIN>;
	    if (($AA_Index>$Prev_AA)&&($AA_Index ne "")&&(($PDB_Chain_ID eq $Protein_Chain_ID)||(($Protein_Chain_ID eq "0")&&($PDB_Chain_ID eq "A"))))
	    {
		$Atoms_Seq[$AA_Index]=$AA_Abb{$AA_Name};
#		print "PDB: $PDB_Chain_ID eq Wanted: $Protein_Chain_ID\n";
#		print "$line\n";
#		print $Atoms_Seq[$AA_Index];
		$Prev_AA=$AA_Index;
	    }
	}
    }
    open (ATOMS_SEQ_FILE,">$Out_File") || die "Can't Open Otput Seq File: $Out_File $!";
    print ATOMS_SEQ_FILE ">$Protein_Name Chain $Protein_Chain_ID ATOMS\n";
    my $i=1;
    my $Seq_Length=@Atoms_Seq;
#    print "\n";
    for ($i=1;$i<$Seq_Length;$i++)
    {
	if ($Atoms_Seq[$i])
	{
	    print ATOMS_SEQ_FILE $Atoms_Seq[$i];
#	    print $Atoms_Seq[$i];
	}
	else
	{
	    print ATOMS_SEQ_FILE "X";
	    $Atoms_Seq[$i]="X";
	    
	}
    }
    close (ATOMS_SEQ_FILE);
    close (Protein_PDB);
    return @Atoms_Seq ;
}

sub Builed_Interactions_Hash_CSU_of_ADD_CSU_TO_PREDICTION_FILE
{
#####        CONTACT MAP FORMAT     ###########
###############################################
#Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance Residue_A_Name Residue_B_Name
#70 18 0 0 1.1 6.1 T E
#135 111 0 0 13.0 4.1 L I

    my $ContactMap_Pos1_Col=0;
    my $ContactMap_Pos1_Chain_Col=2;
    my $ContactMap_Pos1_AA_Col=6;
    my $ContactMap_Pos2_Col=1;
    my $ContactMap_Pos2_Chain_Col=3;
    my $ContactMap_Pos2_AA_Col=7;
    my $Contact_Map_Distance_Col=5;
    
#$Inter_Contact_Map_File,$Intra_Contact_Map_File,$Inter_Intra_Contact_Map_File,$PDB_Chain_ID,$Refference_Seq_PDB_Atoms_Gap);
    my $Contact_Map_File=$_[0];
    my $Protein_Chain_ID=$_[1];
    my $Gap=$_[2];
    my %Interactions_Hash=();
    open (CONTACT_MAP,$Contact_Map_File) || die "Can't Open The Contact Map File $Contact_Map_File $!";
    my $line=<CONTACT_MAP>; #Header
    $line=<CONTACT_MAP>; #Header
    $line=<CONTACT_MAP>; #Header
#
#Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance
#16 83 A A 16.5 2.9
#16 83 B B 15.6 2.9
    
    while ($line=<CONTACT_MAP>)
    {
	my @ContactMap_line_array=split (/\s/,$line);
	my $Residue_A=$ContactMap_line_array[$ContactMap_Pos1_Col];
	my $Residue_B=$ContactMap_line_array[$ContactMap_Pos2_Col];
	my $Chain_of_A=$ContactMap_line_array[$ContactMap_Pos1_Chain_Col];
    	my $Chain_of_B=$ContactMap_line_array[$ContactMap_Pos2_Chain_Col];
	my $Residue_A_and_Gap=$Residue_A+$Gap;
	my $Residue_B_and_Gap=$Residue_B+$Gap;
	#if (($Chain_of_A eq $Protein_Chain_ID) && ($Chain_of_B eq $Protein_Chain_ID))
	#  {
	    $Interactions_Hash{$Residue_A_and_Gap}{$Residue_B_and_Gap}=[@ContactMap_line_array];
	#  }
      }
    close (CONTACT_MAP);
    return \%Interactions_Hash;
  }

sub gap_blast2seq_of_ADD_CSU_TO_PREDICTION_FILE
  {
    my $Protein_seq=$_[0];
    my $Protein_Reff_Seq=$_[1];
    my $Blast2seq_Output=$_[2];
    my $blast2seq_command=WMC_CONSTANTS::BL2SEQ." -i $Protein_seq -j $Protein_Reff_Seq -p blastp -o $Blast2seq_Output -F f";
    print "The Command is: $blast2seq_command\n";
    system ($blast2seq_command);
#  open (BL2SEQ_OUTPUT,"_Atoms__bl2seq1") or die "Can't open BL2Seq Output $!\n";	
    open (BL2SEQ_OUTPUT,$Blast2seq_Output) or die "Can't open BL2Seq Output $!\n";
    my $first_Query=0; #until the first time Query and Sbjct are seen first==0
    my $Sequences_Gap;
    my $Segment_Start=0;
    my $Last_Line_Segment_Start=0;
    my $Segment_End=0;
    my $Hit=0;
    my @Gap_Start_End;
    while (my $line=<BL2SEQ_OUTPUT>)
    {
	if ($line=~/Score/)
	{
	    $Hit=$Hit+1;
	}
	if (($line=~/^Query:/)&&($first_Query==0))
	{
#	    print "$Hit\n";
	    $first_Query=1;
#	    print $line;
	    my @line=split (/\s+/,$line);
	    my $First_PDB_SEQ_AA=$line[1];
	    $line=<BL2SEQ_OUTPUT>;
	    $line=<BL2SEQ_OUTPUT>;
	    @line=split (/\s+/,$line);
	    my $First_PDB_AA_On_Refference_Seq=$line[1];
#	    print $line;
#	    print "First PDB SEQ AA: $First_PDB_SEQ_AA\n";
#	    print "First PDB SEQ AA On Refference Seq: $First_PDB_AA_On_Refference_Seq\n";
	    $Sequences_Gap=$First_PDB_AA_On_Refference_Seq-$First_PDB_SEQ_AA;
	    $Segment_Start=$First_PDB_AA_On_Refference_Seq;
	    $Segment_End=$line[3];
	    $Last_Line_Segment_Start=$line[1];
#	    print "Segment_Start:$Segment_Start\tSegment_End:$Segment_End\n";#<STDIN>;
	}
	if (($line=~/^Sbjct:/)&&($first_Query==1)&&($Hit==1))
	{
	    my @line=split (/\s+/,$line);
	    if ($line[3]>$Segment_End)
	    {
		$Segment_End=$line[3];
#		 print "Segment_End: $Segment_End\n";
	    }
	}
    }
    #system ("rm -r $tmp_file");
    close (BL2SEQ_OUTPUT);
    print "Start=$Segment_Start End=$Segment_End\n";#<STDIN>;
    #    print "Fron Func: Gap=$Sequences_Gap Start=$Segment_Start End=$Segment_End\n";
    my @Return_Items=($Sequences_Gap, $Segment_Start, $Segment_End);
    my %Gap_Translate_Hash=();
#  Score =  403 bits (1036), Expect = e-117
#  Identities = 210/226 (92%), Positives = 210/226 (92%), Gaps = 14/226 (6%)
#
#Query: 1   PRPLLSPPETEEQLLAQAQQLSGYTLGELAALVGLVTPENLKRDKGWIGVLLEIWLGA-- 58
#           PRPLLSPPETEEQLLAQAQQLSGYTLGELAALVGLVTPENLKRDKGWIGVLLEIWLGA
#Sbjct: 4   PRPLLSPPETEEQLLAQAQQLSGYTLGELAALVGLVTPENLKRDKGWIGVLLEIWLGASA 63
#
#Query: 59  ---PEQDFAALGVELKTIPVDSLGRPLETTFVCVAPLTGNSGVTWETSHVRHKLKRVLWI 115
#01234567890   PEQDFAALGVELKTIPVDSLGRPLETTFVCVAPLTGNSGVTWETSHVRHKLKRVLWI
#Sbjct: 64  GSKPEQDFAALGVELKTIPVDSLGRPLETTFVCVAPLTGNSGVTWETSHVRHKLKRVLWI 123
#012345678901234567890123456789012345678901234567890123456789012345678901234
#          1         2         3         4         5         6         7  
#Query: 116 PVEGEASIPLAQRRVGSPLLWSPNEEEDRQLREDWEEL--DIVLGQVERITARHGEYLQI 173
#           PVEGE SIPLAQRRVGSPLLWSPNEEEDRQLREDWEEL   IVLGQVERITARHGEYLQI
#Sbjct: 124 PVEGERSIPLAQRRVGSPLLWSPNEEEDRQLREDWEELMDMIVLGQVERITARHGEYLQI 183
#           123456789012345678901234567890123456789012345678901234567890
#Query: 174 RP-------LTEAIGARGERILTLPRGFYLKKNFTSALLARHFLIQ 212
#           RP       LTEAIGARGERILTLPRGFYLKKNFTSALLARHFLIQ
#Sbjct: 184 RPKAANAKALTEAIGARGERILTLPRGFYLKKNFTSALLARHFLIQ 229   

    open (BL2SEQ_OUTPUT,$Blast2seq_Output) or die "Can't open BL2Seq Output $!\n";
    $first_Query=0; #until the first time Query and Sbjct are seen first==0
    $Hit=0;
    while (my $line=<BL2SEQ_OUTPUT>)
    {
	if ($line=~/Score/)
	{
	    $Hit++;
	}
	if (($line=~/^Query:/) and ($Hit==1))
	{
#	    print "$Hit\n";
	    $first_Query=1;
#	    print $line;
	    my @PDB_line=split (/\s+/,$line);
	    my $PDB_Index=$PDB_line[1];
	    @PDB_line=split (//,$line);
	    $line=<BL2SEQ_OUTPUT>;
	    my @Alingment_line=split (//,$line);
	    $line=<BL2SEQ_OUTPUT>;
	    my @Refference_line=split (/\s+/,$line);
	    my $Reff_Index=$Refference_line[1];
	    @Refference_line=split (//,$line);
	    my $alingment_line_length=@Alingment_line;
	    my $line_length=71;
	    my $counter=11;
	    my $Gap_Counter=$Reff_Index-$PDB_Index;
#	    print "PDB_LINE: @PDB_line\nALINGMENT: @Alingment_line\n REFFERENCE: @Refference_line";<STDIN>;
	    for ($counter;$counter<$line_length;$counter++)
	    {
		if ($counter<=$alingment_line_length)
		{
		    if($PDB_line[$counter] =~ /[A-Z]/)
			#		if($Alingment_line[$counter] =~ /[A-Z ]/)
		    {
			$Reff_Index=$PDB_Index+$Gap_Counter;
			  #		    $Gap_Translate_Hash{$PDB_Index}=$Reff_Index;
			my @tmp_array=($Reff_Index,$Alingment_line[$counter],$Refference_line[$counter],$PDB_line[$counter]);
			#		    $Gap_Translate_Hash{$PDB_Index}="$Reff_Index:$Alingment_line[$counter]*";
			$Gap_Translate_Hash{$PDB_Index}=[@tmp_array];
		    }
		    elsif($PDB_line[$counter] eq "-")
			#		elsif($Alingment_line[$counter] eq "-")
		    {
			$Gap_Counter++;
			$Reff_Index=$PDB_Index+$Gap_Counter;
			my @tmp_array=($Reff_Index,$Alingment_line[$counter],$Refference_line[$counter],$PDB_line[$counter]);
			# $Gap_Translate_Hash{$PDB_Index}=$Reff_Index;
			# $Gap_Translate_Hash{$PDB_Index}="$Reff_Index:$Alingment_line[$counter]*";
			$Gap_Translate_Hash{$PDB_Index}=[@tmp_array];
		    }
		    elsif($Refference_line[$counter] eq "-")
			#		elsif($Alingment_line[$counter] eq "-")
		    {
			$Gap_Counter--;
			$Reff_Index=$PDB_Index+$Gap_Counter;
			my @tmp_array=($Reff_Index,$Alingment_line[$counter],$Refference_line[$counter],$PDB_line[$counter]); 
			# $Gap_Translate_Hash{$PDB_Index}=$Reff_Index;
			# $Gap_Translate_Hash{$PDB_Index}="$Reff_Index:$Alingment_line[$counter]*";
			$Gap_Translate_Hash{$PDB_Index}=[@tmp_array];
		    }
		    if((defined ($Alingment_line[$counter])) &&($Alingment_line[$counter] eq " "))
		    {
			$Reff_Index=$PDB_Index+$Gap_Counter;
			my @tmp_array=($Reff_Index,$Alingment_line[$counter],$Refference_line[$counter],$PDB_line[$counter]); 
			#		    print "$Reff_Index,$Alingment_line[$counter],$Refference_line[$counter],$PDB_line[$counter]\n";
			#		    print "@tmp_array\n";<STDIN>;
			$Known_MisMatch{$Reff_Index}=[@tmp_array];
#			$Known_MisMatch{$PDB_Index}=[@tmp_array];
		    }
		    if((defined ($Alingment_line[$counter]))&&($Alingment_line[$counter] eq "+"))
		    {
			$Reff_Index=$PDB_Index+$Gap_Counter;
			my @tmp_array=($Reff_Index,$Alingment_line[$counter],$Refference_line[$counter],$PDB_line[$counter]); 
			#		    print "$Reff_Index,$Alingment_line[$counter],$Refference_line[$counter],$PDB_line[$counter]\n";
			#		    print "@tmp_array\n";<STDIN>;
			$Known_MisMatch{$Reff_Index}=[@tmp_array];
#			$Known_MisMatch{$PDB_Index}=[@tmp_array];
		    }
		    
		    $Reff_Index++;
		    $PDB_Index++;
		}
	    }
	}
    }
    
    close (BL2SEQ_OUTPUT);
    unlink $Blast2seq_Output;
    push (@Return_Items,\%Gap_Translate_Hash,\%Known_MisMatch);
    # Print_Hash_Of_Array (\%Gap_Translate_Hash,"Test_Gap_bls2seq_1AZO_0");
    # Print_Hash_Of_Array (\%Known_MisMatch,"Test_MisMatch_1AZO_0");
    return @Return_Items;
}

sub CSU_ContactMap_MMOL_Only_With_AA_Inter_Chain_Intra_Chain_Interactions1
{
    
    my ($Protein_File_Name, $OutputFile_Sur,$Chain_ID);
    my $intra_Counter=0;
    my $inter_Counter=0;
    $Protein_File_Name=shift;
    $OutputFile_Sur=shift;
    $Chain_ID=shift;
    
    open (PDB_FILE,$Protein_File_Name)||die "cannot open file:$!";
    open (OUTPUT, ">$OutputFile_Sur") || die "Can't open output file: $OutputFile_Sur $!.\n";
    my $Residue_num;#=1;
    
    my $Prev_AA=0;
    my $Prev_Chain;
    my %Inter_Chain_Interactions=();
    my %Intra_Chain_Interactions=();
    my %Both_Inter_Intra_Chain_Interactions=();
    my $Interacted_AA_Name;
    my $Interacted_AA_Num;
    my $Interacted_AA_Dis;
    my $Intearcted_AA_Surf;
    my $Interacted_AA_Chain;
    my $Chain_num;
     my %AA_Names=(GLY=>'1',ALA=>'1',VAL=>'1',LEU=>'1',ILE=>'1',MET=>'1',PHE=>'1',TRP=>'1',PRO=>'1',SER=>'1',THR=>'1',CYS=>'1',
		  TYR=>'1',ASN=>'1',GLN=>'1',ASP=>'1',GLU=>'1',LYS=>'1',ARG=>'1',HIS=>'1');
    my %AA_Abbreviations=(GLY=>'G',ALA=>'A',VAL=>'V',LEU=>'L',ILE=>'I',MET=>'M',PHE=>'F',TRP=>'W',PRO=>'P',SER=>'S',THR=>'T',CYS=>'C',TYR=>'Y',ASN=>'N',GLN=>'Q',ASP=>'D',GLU=>'E',LYS=>'K',ARG=>'R',HIS=>'H');
    print OUTPUT "Intra Chain Interactions at: $Protein_File_Name\n\n";
    print OUTPUT "Residue_A Residue_B Chain_of_A Chain_of_B Surface Distance Residue_A_Name Residue_B_Name\n";
    
###Getting the information about $Total_Residues and $Chain_num from The PDB File  ###
    while (my $line=<PDB_FILE>)
    {
	if ($line=~ /^ATOM/)
	{
#	print "ATOM LINE\n";<STDIN>;
	    if ($line=~ /^.{21}([A-Z ]{1})([0-9 ]{4})/)
	    {
		$Chain_num=$1;
		$Residue_num=$2;
	    }
	    $Chain_num=remove_whitespace($Chain_num);
	    $Residue_num=remove_whitespace($Residue_num);
#	print "Chain:$Chain_num\tPos:$Residue_num\n";#<STDIN>;
	    
	    
### Calling CSU for the Residue of specific Chain in order to find who interacts with him
	    if (($Residue_num>$Prev_AA)||($Chain_num gt $Prev_Chain))
	    {
		my $CSU_Command=WMC_CONSTANTS::CSU." $Protein_File_Name $Residue_num $Chain_num |";
		open (README, $CSU_Command) or die "Can't Run ProgramL $!\n";
		my $Beg_Flag=0;
		my $Data;
		my $AA1_Name="";
		while (<README>)
		{
		    if ($_=~ /Residues in contact with ([A-Z]{3})/)
		    {
			$AA1_Name=$1;
		    }
		    if ($_=~m/DC$/)
		    {
			$Beg_Flag=1;
		    }
		    if (($Beg_Flag==1)&&($_=~m/^------------------/))
		    {
			$Beg_Flag=2;
		    }
		    if ($_=~ m/Table III$/)
		    {
			$Beg_Flag=0;
		    }
		    if ($Beg_Flag==2)
		    {
			$Data=1;
			if (($_=~m/^------------------/)||($_=~m/\t+$/))
			{
			    $Data=0;
			}
			if ($Data==1)
			{
			    my @line_array=split(/\s+/,$_);
			    chomp ($_);
			    if ($_ ne "")
			    {
				$Interacted_AA_Name=$line_array[2];
				$Interacted_AA_Num=$line_array[1];
				$Interacted_AA_Dis=$line_array[3];
				$Intearcted_AA_Surf=$line_array[4];
				if ($line_array[1]=~m/^(\d+)/)
				{
				    $Interacted_AA_Num=$1;
				    $Interacted_AA_Chain="";
				}
				if ($line_array[1]=~m/^(\d+)([A-Z])/)
				{
				    $Interacted_AA_Num=$1;
				    $Interacted_AA_Chain=$2;
				}
				if  ($Interacted_AA_Name=~/([A-Z]*)\*/)
				{
				    $Interacted_AA_Name=$1;
				}
				my $Pair=$Residue_num."_".$Interacted_AA_Num;
				my $Chain_num_For_File;
				if ($Chain_num eq "")
				{
				    $Chain_num_For_File=0;
				}
				else
				{
				    $Chain_num_For_File=$Chain_num;
				}
				my $Interacted_AA_Chain_For_File;
				if ($Interacted_AA_Chain eq "")
				{
				    $Interacted_AA_Chain_For_File=0;
				}
				else
				{
				    $Interacted_AA_Chain_For_File=$Interacted_AA_Chain;
				}
				
				my $Conatact_Map_Data;
				if (($AA_Abbreviations{$AA1_Name})&&($AA_Abbreviations{$Interacted_AA_Name})) # Only Natural AA
				{
				    $Conatact_Map_Data="$Residue_num $Interacted_AA_Num $Chain_num_For_File $Interacted_AA_Chain_For_File $Intearcted_AA_Surf $Interacted_AA_Dis $AA_Abbreviations{$AA1_Name} $AA_Abbreviations{$Interacted_AA_Name}"; 
				}
				if (($AA_Names{$Interacted_AA_Name}&&($Interacted_AA_Chain eq $Chain_num)&&(($Interacted_AA_Chain eq $Chain_ID)||(($Interacted_AA_Chain eq "")&&($Chain_ID eq "0"))||(($Interacted_AA_Chain eq "A")&&($Chain_ID eq "0"))))) # Choosing only those who are AA Interaction and are IntraChain interactions
				{
				    print OUTPUT "$Conatact_Map_Data\n";
				}
			    }
			    
			}
		    }
		}
	    }
	    close (README);
	    $Prev_AA=$Residue_num;
	    $Prev_Chain=$Chain_num;
	}
    }
    close (OUTPUT);
    close (PDB_FILE);
}

1;
